//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "FOStationWrappers.h"
#include "ODBCWrappers.h"
#include "WinHTTPWrappers.h"
#include "VersionInfo.h"
#include "ntserv.h"
#include "YKUSHWrappers.h"
#include "SQLiteWrappers.h"
#include "DTime+JulianCalendar.h"
#include "DTime+GregorianCalendar.h"
#include "DTime+HebrewCalendar.h"
#include "DTime+CDate.h"
#include "DTime+CLDate.h"


//////////////////// Implementation ///////////////////////////////////////////

//The NT service itself
class CMyService : public CNTService
{
public:
//Constructors / Destructors
  CMyService();
  CMyService(const CMyService&) = delete;
  CMyService(CMyService&&) = delete;
  ~CMyService() = default; //NOLINT(modernize-use-override)

//Methods
  CMyService& operator=(const CMyService&) = delete;
  CMyService& operator=(CMyService&&) = delete;
  static double DewPoint(_In_ double fTemperature, _In_ double fRelativeHumidity) noexcept;

protected:
  enum class FOStationType
  {
    WithSolarData = 0,
    WithourSolarData = 1
  };

//Methods
  void WINAPI ServiceMain(_In_ DWORD dwArgc, _In_opt_ LPTSTR* pszArgv) override;
  void OnStop() noexcept override;
  void OnShutdown() noexcept override;
  void ShowHelp() noexcept override;
  bool SynchronizeWithStation();
  void ProcessHistoricalRecord(_Inout_ CFORecord& record);
  void ProcessCurrentRecord(_Inout_ CFORecord& record);
  void LogRawRecord(_In_z_ LPCTSTR pszPrefix, _In_ const CFORecord& record);
  void LogRawLightRecord(_In_z_ LPCTSTR pszPrefix, _In_ const CFORecord& record);
  bool ReadHistoricalData();
  bool LogODBC(_In_ const CODBC::CHandle& handle, _In_z_ const TCHAR* pszFormat, ...);
  bool LogWin32Error(_In_ DWORD dwLastError, _In_z_ const TCHAR* pszFormat, ...);
  bool Log(_In_z_ const TCHAR* pszFormat, ...);
  bool OpenDB(_In_ bool bAttemptSchemaCreation);
  bool LogRecordToDB(_In_ CFORecord& record);
  bool SafeToReadFromDevice() noexcept;
  bool UploadToWOW(_In_ const CFORecord& record);
  bool UploadToWeatherUnderground(_In_ const CFORecord& record);
  void CalculateRainfallValue(_Inout_ CFORecord& record);
  bool ResetUSBWithYKUSHXS();
  bool OpenWeatherStationAsHIDDevice();
  void ReadCOMPortData(_Inout_ ATL::CStringA& sBuffer, _Inout_ OVERLAPPED& overlapped, _Inout_ CFORecord& lastWeatherRecord, _Inout_ CFORecord& lastLightRecord);
  void HandleCOMPortData(_In_z_ LPCSTR pszNewData, _Inout_ ATL::CStringA& sBuffer, _Inout_ CFORecord& lastWeatherRecord, _Inout_ CFORecord& lastLightRecord);

//Member variables
  CFOStation m_Device; //The weather station device
  ATL::CEvent m_StopEvent; //Event to request a stop
  FOStationType m_StationType; //The type of the station type
  ATL::CString m_sLogFilenameSpecifier; //The filename specifier for the log file
  ATL::CString m_sCurrentLogFilename; //The current filename we are logging to
  ATL::CString m_sLogDirectory; //The log directory we are using
  ATL::CAtlFile m_Log; //The log file we will be using
  DTimePlus::CLDate<DTimePlus::CGregorianCalendar> m_StationClock; //The local time when the current write address changed while synchronizing with the station
  DTimePlus::CLDate<DTimePlus::CGregorianCalendar> m_SensorClock; //The local time when the contents of the current write address changed while synchronizing with the station
  CODBC::CConnection m_DBConnection; //Our connection to the database
  SQLite3::CDB m_SqliteDB; //The Sqlite DB we are using
  ATL::CString m_sDBInitString; //The init string to use to connect to the DB
  ATL::CString m_sSqliteDatabase; //The path to the Sqlite3 database file to use.
  int m_nAvoidReadPeriod; //The period in seconds which we use to avoid reading from the station when the station is writing date
  UINT m_nMaxHistoricalRecords; //The maximum number of historical records to process
  DWORD m_dwDeviceTimeout; //The timeout in ms to use for reading / writing to the weather station
  ATL::CString m_sMETWOWSiteID; //The Met Office WOW site id to use
  ATL::CString m_sMETWOWSiteAuthenticationKey; //The Met Office WOW site authentication key to use
  ATL::CString m_sWeatherUndergroundID; //The Weather Underground ID to use
  ATL::CString m_sWeatherUndergroundPassword; //The Weather Underground password to use
  bool m_bDebugAuditing; //Should we include additional "debug" auditing
  bool m_bResetUsingYKUSHXS; //Should we reset the  YKUSH XS when we detect USB lockups?
  HID::CYKUSHXSDevice m_YKUSHFXDevice; //The CYKUSHXSDevice we use to handle USB lockups
  HID::String m_sDevicePath; //The path used to open the weather station
  int m_nMETWOWLoggingInterval; //The interval in minutes to log data to the MET WOW web site
  int m_nWeatherUndergroundLoggingInterval; //The interval in minutes to log data to the Weather Underground web site
  int m_nDBLoggingInterval; //The interval in minutes to log data to the database
  DTimePlus::CLDate<DTimePlus::CGregorianCalendar> m_LastMETWOWUpload; //The last UTC time a MEM WOW upload has occurred
  DTimePlus::CLDate<DTimePlus::CGregorianCalendar> m_LastWeatherUndergroundUpload; //The last UTC time a Weather Underground upload has occurred
  DTimePlus::CLDate<DTimePlus::CGregorianCalendar> m_LastDBSave; //The last UTC time a DB update occurred for current records;
  int m_nMidnightRainfallCount; //The rainfall count at the last midnight
  int m_nLastRainfallCount; //The last rainfall count logged to the database
  DTimePlus::CLDate<DTimePlus::CGregorianCalendar> m_LastRainfallCount; //The UTC date / time associated with "m_nLastRainfallCount"
  int m_nYearMidnightRainfall; //The year part corresponding to "m_nMidnightRainfallCount"
  int m_nMonthMidnightRainfall; //The month part corresponding to "m_nMidnightRainfallCount"
  int m_nDayMidnightRainfall; //The day part corresponding to "m_nMidnightRainfallCount"
  ATL::CString m_sVersionInfo; //The version info we use when calling the MET WOW service
  int m_nCOMPort; //The COM port to open if we are using the Arduino RF Receiver unit instead of the original USB interface
  int m_nBaudRate; //The baud rate to open the serial port at
  ATL::CEvent m_COMPortDataEvent; //The CEvent associated with the COM port
  ATL::CHandle m_Port; //The COM port
};


#pragma warning(suppress: 26426)
CMyService theService;


#pragma warning(suppress: 26455)
CMyService::CMyService() : CNTService{_T("FOSTATION"), _T("Fine Offset Weather Station"), SERVICE_ACCEPT_STOP | SERVICE_ACCEPT_SHUTDOWN, _T("Logs data from Fine Offset compatible Weather Stations")},
                           m_StopEvent{TRUE, FALSE},
                           m_StationType{FOStationType::WithSolarData},
                           m_sLogFilenameSpecifier{_T("FOStation.log")},
                           m_sDBInitString{_T("DRIVER={ODBC Driver 13 for SQL Server};SERVER=.;DATABASE=WeatherStation;Trusted_Connection=yes;")}, //By default use a database on the local machine called "WeatherStation"
                           m_nAvoidReadPeriod{3},
                           m_nMaxHistoricalRecords{0},
                           m_dwDeviceTimeout{8000},
                           m_bDebugAuditing{false},
                           m_bResetUsingYKUSHXS{false},
                           m_nMETWOWLoggingInterval{15},
                           m_nWeatherUndergroundLoggingInterval{15},
                           m_nDBLoggingInterval{5},
                           m_nMidnightRainfallCount{INT_MIN},
                           m_nLastRainfallCount{-1},
                           m_nYearMidnightRainfall{-1},
                           m_nMonthMidnightRainfall{-1},
                           m_nDayMidnightRainfall{-1},
                           m_sVersionInfo{_T("1.0")},
                           m_nCOMPort{0},
                           m_nBaudRate{115200},
                           m_COMPortDataEvent{TRUE, FALSE}
{
}

void CMyService::CalculateRainfallValue(_Inout_ CFORecord& record)
{
  //Handle overflow of the rainfall counter value
  if (record.m_nTotalRain.has_value() && (m_nLastRainfallCount != -1) && (record.m_nTotalRain.value() < m_nLastRainfallCount))
  {
    record.m_bRainfallOverflowFlag = true;

    //PJ TO DO: To do email notification on rainfall overflow here
    Log(_T("Total Rain value has overflowed to %d\n"), record.m_nTotalRain.value());

    //Patch up the midnight rainfall values in the registry upon overflow
    m_nMidnightRainfallCount = record.m_nTotalRain.value() - (m_nLastRainfallCount - m_nMidnightRainfallCount);
    const auto gd = record.m_LocalDateTime.GetDate();
    m_nYearMidnightRainfall = gd.Year;
    m_nMonthMidnightRainfall = gd.Month;
    m_nDayMidnightRainfall = gd.Day;
    static const TCHAR* pszMidnightRainfallSection{_T("MidnightRainfall")};
    WriteProfileInt(pszMidnightRainfallSection, _T("Count"), m_nMidnightRainfallCount);
    WriteProfileInt(pszMidnightRainfallSection, _T("Year"), m_nYearMidnightRainfall);
    WriteProfileInt(pszMidnightRainfallSection, _T("Month"), m_nMonthMidnightRainfall);
    WriteProfileInt(pszMidnightRainfallSection, _T("Day"), m_nDayMidnightRainfall);
    if (m_bDebugAuditing)
      Log(_T("Remembering midnight rainfall of %d for %s\n"), m_nMidnightRainfallCount, record.m_LocalDateTime.Format(_T("%x")).c_str());

    //Update the last rainfall count in the registry upon overflow
    m_nLastRainfallCount = record.m_nTotalRain.value();
    m_LastRainfallCount = record.m_UTCDateTime;
    static const TCHAR* pszLastRainfallSection = _T("LastRainfall");
    WriteProfileInt(pszLastRainfallSection, _T("Count"), m_nLastRainfallCount);
    const auto gd2 = m_LastRainfallCount.GetDate();
    WriteProfileInt(pszLastRainfallSection, _T("Year"), gd2.Year);
    WriteProfileInt(pszLastRainfallSection, _T("Month"), gd2.Month);
    WriteProfileInt(pszLastRainfallSection, _T("Day"), gd2.Day);
    WriteProfileInt(pszLastRainfallSection, _T("Hour"), gd2.Hour);
    WriteProfileInt(pszLastRainfallSection, _T("Minute"), gd2.Minute);
    WriteProfileInt(pszLastRainfallSection, _T("Second"), gd2.Second);
    if (m_bDebugAuditing)
      Log(_T("Updating last rainfall count of %d for %s\n"), m_nLastRainfallCount, record.m_LocalDateTime.Format(_T("%x")).c_str());
  }

  //Calculate the rainfall rate since the last rainfall count if we can and disregard it if the rainfall rate is abnormally high
  if (record.m_nTotalRain.has_value() && (m_nLastRainfallCount != -1) && 
      (m_LastRainfallCount.IsValid()) && (record.m_UTCDateTime != m_LastRainfallCount))
  {
    const auto rainfallRate((static_cast<double>(record.m_nTotalRain.value()) - m_nLastRainfallCount) / (record.m_UTCDateTime - m_LastRainfallCount).SecondsAsDouble() * 3600);
    if (rainfallRate > 1000) //The world record for the most rain in one hour is 300 mm so anything greater than that we will consider as invalid and ignore the rainfall count in this record
    {
      //PJ TO DO: To do email notification on rainfall rate being too great a change to be valid
      Log(_T("Modified Total Rain value from %d to NULL as it is too great a change to be valid\n"), record.m_nTotalRain.value());
      record.m_nTotalRain.reset();
    }
  }

  //Determine the rainfall in mm since the last record
  if (record.m_nTotalRain.has_value() && (m_nLastRainfallCount != -1))
  {
    record.m_fRainfall = static_cast<float>((record.m_nTotalRain.value() - m_nLastRainfallCount) * 0.3);
    if (m_bDebugAuditing)
      Log(_T("Calculating rainfall to be %f\n"), record.m_fRainfall.value());
  }

  //Determine the rainfall in mm since midnight
  if (record.m_nTotalRain.has_value() && (m_nMidnightRainfallCount != INT_MIN))
  {
    record.m_fRainfallSinceMidnight = static_cast<float>((record.m_nTotalRain.value() - m_nMidnightRainfallCount) * 0.3);
    if (m_bDebugAuditing)
      Log(_T("Calculating rainfall since midnight to be %f\n"), record.m_fRainfallSinceMidnight.value());
  }

  //Remember the midnight rainfall count and date if necessary
  if (record.m_nTotalRain.has_value())
  {
    const auto gd = record.m_LocalDateTime.GetDate();
    const int nNewYear{gd.Year};
    const int nNewMonth{gd.Month};
    const int nNewDay{gd.Day};
    if (((m_nYearMidnightRainfall == -1) || (m_nMonthMidnightRainfall == -1) || (m_nDayMidnightRainfall == -1)) ||
        ((nNewYear != m_nYearMidnightRainfall) || (nNewMonth != m_nMonthMidnightRainfall) || (nNewDay != m_nDayMidnightRainfall)))
    {
      m_nMidnightRainfallCount = record.m_nTotalRain.value();
      m_nYearMidnightRainfall = nNewYear;
      m_nMonthMidnightRainfall = nNewMonth;
      m_nDayMidnightRainfall = nNewDay;

      //Save the midnight rainfall values to allow us to reuse these values on service restart
      static const TCHAR* pszMidnightRainfallSection{_T("MidnightRainfall")};
      WriteProfileInt(pszMidnightRainfallSection, _T("Count"), m_nMidnightRainfallCount);
      WriteProfileInt(pszMidnightRainfallSection, _T("Year"), m_nYearMidnightRainfall);
      WriteProfileInt(pszMidnightRainfallSection, _T("Month"), m_nMonthMidnightRainfall);
      WriteProfileInt(pszMidnightRainfallSection, _T("Day"), m_nDayMidnightRainfall);

      if (m_bDebugAuditing)
        Log(_T("Remembering midnight rainfall of %d for beginning of %s\n"), m_nMidnightRainfallCount, record.m_LocalDateTime.Format(_T("%x")).c_str());
    }
  }
}

void CMyService::ProcessHistoricalRecord(_Inout_ CFORecord& record)
{
  if (m_bDebugAuditing)
    Log(_T("Processing historical record for %s\n"), record.m_LocalDateTime.Format(_T("%x")).c_str());

  //Calculate the rainfall values
  CalculateRainfallValue(record);

  if (m_bDebugAuditing)
    LogRawRecord(_T("Historical record: %s\n"), record);

  //PJ TO DO: To do email notification on record.m_bRainfallOverflowFlag == True
  //PJ TO DO: To do email notification on record.m_bBatteryStatusOK == False
  //PJ TO DO: To do email notification on record.m_bLightStatusOK == False

  //Store the record in the database
  LogRecordToDB(record);
}

void CMyService::LogRawRecord(_In_z_ LPCTSTR pszPrefix, _In_ const CFORecord& record)
{
  //Log the raw record details
  ATL::CString sRawRecord;
  if (record.m_nIndoorRelativeHumidity.has_value())
    sRawRecord.Format(_T("HumidityI:%d"), record.m_nIndoorRelativeHumidity.value());
  else
    sRawRecord = _T("HumidityI:NULL");
  if (record.m_fIndoorTemperature.has_value())
    sRawRecord.AppendFormat(_T(", TempI:%.2f"), record.m_fIndoorTemperature.value());
  else
    sRawRecord += _T(", TempI:NULL");
  if (record.m_nOutdoorRelativeHumidity.has_value())
    sRawRecord.AppendFormat(_T(", HumidityO:%d"), record.m_nOutdoorRelativeHumidity.value());
  else
    sRawRecord += _T(", HumidityO:NULL");
  if (record.m_fOutdoorTemperature.has_value())
    sRawRecord.AppendFormat(_T(", TempO:%.2f"), record.m_fOutdoorTemperature.value());
  else
    sRawRecord += _T(", TempO:NULL");
  if (record.m_fAbsolutePressure.has_value())
    sRawRecord.AppendFormat(_T(", Pressure:%.2f"), record.m_fAbsolutePressure.value());
  else
    sRawRecord += _T(", Pressure:NULL");
  if (record.m_fAverageWindSpeed.has_value())
    sRawRecord.AppendFormat(_T(", AvgW:%.2f"), record.m_fAverageWindSpeed.value());
  else
    sRawRecord += _T(", AvgW:NULL");
  if (record.m_fGustWindSpeed.has_value())
    sRawRecord.AppendFormat(_T(", GustW:%.2f"), record.m_fGustWindSpeed.value());
  else
    sRawRecord += _T(", GustW:NULL");
  if (record.m_nWindDirection.has_value())
    sRawRecord.AppendFormat(_T(", DirW:%d"), record.m_nWindDirection.value());
  else
    sRawRecord += _T(", DirW:NULL");
  if (record.m_nTotalRain.has_value())
    sRawRecord.AppendFormat(_T(", Rain:%d"), record.m_nTotalRain.value());
  else
    sRawRecord += _T(", Rain:NULL");
  if (record.m_bRainfallOverflowFlag)
    sRawRecord += _T(", RainOverflow:1");
  else
    sRawRecord += _T(", RainOverflow:0");
  if (record.m_fIlluminance.has_value())
    sRawRecord.AppendFormat(_T(", Lux:%.2f"), record.m_fIlluminance.value());
  else
    sRawRecord += _T(", Lux:NULL");
  if (record.m_nUV.has_value())
    sRawRecord.AppendFormat(_T(", UV:%d"), record.m_nUV.value());
  else
    sRawRecord += _T(", UV:NULL");
  if (record.m_fRainfallSinceMidnight.has_value())
    sRawRecord.AppendFormat(_T(", RainMidnight:%.1f"), record.m_fRainfallSinceMidnight.value());
  else
    sRawRecord += _T(", RainMidnight:NULL");
  if (record.m_fRainfall.has_value())
    sRawRecord.AppendFormat(_T(", RainMM:%.1f"), record.m_fRainfall.value());
  else
    sRawRecord += _T(", RainMM:NULL");
  if (record.m_bBatteryStatusOK.has_value())
  {
    if (record.m_bBatteryStatusOK.value())
      sRawRecord += _T(", BatteryStatus:1");
    else
      sRawRecord += _T(", BatteryStatus:0");
  }
  else
    sRawRecord += _T(", BatteryStatus:NULL");
  if (record.m_nRSSI.has_value())
    sRawRecord.AppendFormat(_T(", RSSI:%d"), record.m_nRSSI.value());
  else
    sRawRecord += _T(", RSSI:NULL");
  if (record.m_nSensorID.has_value())
    sRawRecord.AppendFormat(_T(", SensorID:%d"), record.m_nSensorID.value());
  else
    sRawRecord += _T(", SensorID:NULL");
  if (record.m_bLightStatusOK.has_value())
  {
    if (record.m_bLightStatusOK.value())
      sRawRecord += _T(", SensorStatus:1");
    else
      sRawRecord += _T(", SensorStatus:0");
  }
  else
    sRawRecord += _T(", SensorStatus:NULL");

  Log(pszPrefix, sRawRecord.GetString());
}

void CMyService::LogRawLightRecord(_In_z_ LPCTSTR pszPrefix, _In_ const CFORecord& record)
{
  //Log the raw record details
  ATL::CString sRawRecord;
  if (record.m_nIndoorRelativeHumidity.has_value())
    sRawRecord.Format(_T("HumidityI:%d"), record.m_nIndoorRelativeHumidity.value());
  else
    sRawRecord = _T("HumidityI:NULL");
  if (record.m_fIndoorTemperature.has_value())
    sRawRecord.AppendFormat(_T(", TempI:%.2f"), record.m_fIndoorTemperature.value());
  else
    sRawRecord += _T(", TempI:NULL");
  if (record.m_fAbsolutePressure.has_value())
    sRawRecord.AppendFormat(_T(", Pressure:%.2f"), record.m_fAbsolutePressure.value());
  else
    sRawRecord += _T(", Pressure:NULL");
  if (record.m_fIlluminance.has_value())
    sRawRecord.AppendFormat(_T(", Lux:%.2f"), record.m_fIlluminance.value());
  else
    sRawRecord += _T(", Lux:NULL");
  if (record.m_nUV.has_value())
    sRawRecord.AppendFormat(_T(", UV:%d"), record.m_nUV.value());
  else
    sRawRecord += _T(", UV:NULL");
  if (record.m_nRSSI.has_value())
    sRawRecord.AppendFormat(_T(", RSSI:%d"), record.m_nRSSI.value());
  else
    sRawRecord += _T(", RSSI:NULL");
  if (record.m_nSensorID.has_value())
    sRawRecord.AppendFormat(_T(", SensorID:%d"), record.m_nSensorID.value());
  else
    sRawRecord += _T(", SensorID:NULL");
  if (record.m_bLightStatusOK.has_value())
  {
    if (record.m_bLightStatusOK.value())
      sRawRecord += _T(", SensorStatus:1");
    else
      sRawRecord += _T(", SensorStatus:0");
  }
  else
    sRawRecord += _T(", SensorStatus:NULL");

  Log(pszPrefix, sRawRecord.GetString());
}

void CMyService::ProcessCurrentRecord(_Inout_ CFORecord& record)
{
  //Calculate the rainfall values
  CalculateRainfallValue(record);

  if (m_bDebugAuditing)
    LogRawRecord(_T("Current Record: %s\n"), record);

  //Store the record in the database if required
  LogRecordToDB(record);

  //Upload to the MET WOW web site if required
  if (!m_sMETWOWSiteID.IsEmpty() && !m_sMETWOWSiteAuthenticationKey.IsEmpty())
    UploadToWOW(record);

  //Upload to the Weather Underground web site if required
  if (!m_sWeatherUndergroundID.IsEmpty() && !m_sWeatherUndergroundPassword.IsEmpty())
    UploadToWeatherUnderground(record);
}

bool CMyService::LogRecordToDB(_In_ CFORecord& record)
{
  bool bSave{false};
  const DTimePlus::CLDate<DTimePlus::CGregorianCalendar> now{DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentSystemTime()};
  if (m_LastDBSave.IsValid())
  {
    if (m_nDBLoggingInterval == 0)
      bSave = true;
    else
    {
      const DTimePlus::CLTimeSpan timeSpan{now - m_LastDBSave};
      bSave = (timeSpan.SecondsAsDouble() > (m_nDBLoggingInterval * 60.0));
    }
  }
  else
    bSave = true;
  if (!bSave)
    return false;

  if (m_bDebugAuditing)
    Log(_T("Saving record to database\n"));

  //Reopen the DB connection if required
  if (!OpenDB(false))
    return false;

  //Handle the Sqlite code path for DB logging
  if (!m_sSqliteDatabase.IsEmpty())
  {
    //The Sqlite codepath

    SQLite3::CStatement stmt;
    int nRet{stmt.Preparev2(m_SqliteDB, "INSERT INTO historicalData"
                                        "(dateTime, "
                                        "indoorHumidity, "
                                        "indoorTemperature, "
                                        "outdoorHumidity, "
                                        "outdoorTemperature, "
                                        "pressure, "
                                        "avgWindSpeed, "
                                        "gustWindSpeed, "
                                        "windDirection, "
                                        "totalRainfall, "
                                        "rainfallSinceMidnight, "
                                        "rainfall, "
                                        "UV, "
                                        "illuminance)"
                                        " VALUES(?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11, ?12, ?13, ?14) ON CONFLICT(dateTime) "
                                        "DO UPDATE SET "
                                        "dateTime=?1, "
                                        "indoorHumidity=?2, "
                                        "indoorTemperature=?3, "
                                        "outdoorHumidity=?4, "
                                        "outdoorTemperature=?5, "
                                        "pressure=?6, "
                                        "avgWindSpeed=?7, "
                                        "gustWindSpeed=?8, "
                                        "windDirection=?9, "
                                        "totalRainfall=?10, "
                                        "rainfallSinceMidnight=?11, "
                                        "rainfall=?12, "
                                        "UV=?13, "
                                        "illuminance=?14;")};
    if (nRet != SQLITE_OK)
    {
      Log(_T("Failed to prepare statement while logging to database, Error:%d\n"), nRet);
      return false;
    }

    //Bind the parameters
    const double fUnixDateTime{record.m_LocalDateTime.GetValue() - (141427 * 86400.0)}; //141427 corresponds to the GDN of 1 January 1970
    nRet = stmt.Bind(1, fUnixDateTime);
    if (nRet != SQLITE_OK)
    {
      Log(_T("Failed to bind parameter 1 while logging to database, Error:%d\n"), nRet);
      return false;
    }

    if (record.m_nIndoorRelativeHumidity.has_value())
      nRet = stmt.Bind(2, record.m_nIndoorRelativeHumidity.value());
    else
      nRet = stmt.BindNULL(2);
    if (nRet != SQLITE_OK)
    {
      Log(_T("Failed to bind parameter 2 while logging to database, Error:%d\n"), nRet);
      return false;
    }

    if (record.m_fIndoorTemperature.has_value())
      nRet = stmt.Bind(3, record.m_fIndoorTemperature.value());
    else
      nRet = stmt.BindNULL(3);
    if (nRet != SQLITE_OK)
    {
      Log(_T("Failed to bind parameter 3 while logging to database, Error:%d\n"), nRet);
      return false;
    }
    if (record.m_nOutdoorRelativeHumidity.has_value())
      nRet = stmt.Bind(4, record.m_nOutdoorRelativeHumidity.value());
    else
      nRet = stmt.BindNULL(4);
    if (nRet != SQLITE_OK)
    {
      Log(_T("Failed to bind parameter 4 while logging to database, Error:%d\n"), nRet);
      return false;
    }
    if (record.m_fOutdoorTemperature.has_value())
      nRet = stmt.Bind(5, record.m_fOutdoorTemperature.value());
    else
      nRet = stmt.BindNULL(5);
    if (nRet != SQLITE_OK)
    {
      Log(_T("Failed to bind parameter 5 while logging to database, Error:%d\n"), nRet);
      return false;
    }
    if (record.m_fAbsolutePressure.has_value())
      nRet = stmt.Bind(6, record.m_fAbsolutePressure.value());
    else
      nRet = stmt.BindNULL(6);
    if (nRet != SQLITE_OK)
    {
      Log(_T("Failed to bind parameter 6 while logging to database, Error:%d\n"), nRet);
      return false;
    }
    if (record.m_fAverageWindSpeed.has_value())
      nRet = stmt.Bind(7, record.m_fAverageWindSpeed.value());
    else
      nRet = stmt.BindNULL(7);
    if (nRet != SQLITE_OK)
    {
      Log(_T("Failed to bind parameter 7 while logging to database, Error:%d\n"), nRet);
      return false;
    }
    if (record.m_fGustWindSpeed.has_value())
      nRet = stmt.Bind(8, record.m_fGustWindSpeed.value());
    else
      nRet = stmt.BindNULL(8);
    if (nRet != SQLITE_OK)
    {
      Log(_T("Failed to bind parameter 8 while logging to database, Error:%d\n"), nRet);
      return false;
    }
    if (record.m_nWindDirection.has_value())
      nRet = stmt.Bind(9, record.m_nWindDirection.value());
    else
      nRet = stmt.BindNULL(9);
    if (nRet != SQLITE_OK)
    {
      Log(_T("Failed to bind parameter 9 while logging to database, Error:%d\n"), nRet);
      return false;
    }
    if (record.m_nTotalRain.has_value())
      nRet = stmt.Bind(10, record.m_nTotalRain.value() * 0.3);
    else
      nRet = stmt.BindNULL(10);
    if (nRet != SQLITE_OK)
    {
      Log(_T("Failed to bind parameter 10 while logging to database, Error:%d\n"), nRet);
      return false;
    }
    if (record.m_fRainfallSinceMidnight.has_value())
      nRet = stmt.Bind(11, record.m_fRainfallSinceMidnight.value());
    else
      nRet = stmt.BindNULL(11);
    if (nRet != SQLITE_OK)
    {
      Log(_T("Failed to bind parameter 11 while logging to database, Error:%d\n"), nRet);
      return false;
    }
    if (record.m_fRainfall.has_value())
      nRet = stmt.Bind(12, record.m_fRainfall.value());
    else
      nRet = stmt.BindNULL(12);
    if (nRet != SQLITE_OK)
    {
      Log(_T("Failed to bind parameter 12 while logging to database, Error:%d\n"), nRet);
      return false;
    }
    if (record.m_nUV.has_value())
      nRet = stmt.Bind(13, record.m_nUV.value());
    else
      nRet = stmt.BindNULL(13);
    if (nRet != SQLITE_OK)
    {
      Log(_T("Failed to bind parameter 13 while logging to database, Error:%d\n"), nRet);
      return false;
    }
    if (record.m_fIlluminance.has_value())
      nRet = stmt.Bind(14, record.m_fIlluminance.value());
    else
      nRet = stmt.BindNULL(14);
    if (nRet != SQLITE_OK)
    {
      Log(_T("Failed to bind parameter 14 while logging to database, Error:%d\n"), nRet);
      return false;
    }

    nRet = stmt.Step();
    if (nRet != SQLITE_DONE)
    {
      //PJ TO DO: To do email notification on failure here
      Log(_T("Failed to log record to database, Error:%d\n"), nRet);
      return false;
    }
  }
  else
  {
    //The ODBC codepath

    //Form the SQL we will be using
    ATL::CString sSQL{_T("{ CALL historicalDataUpsert(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?) }")};

    //Create and prepare the statement
    CODBC::CStatement kODBCStatement;
    SQLRETURN nRet{kODBCStatement.Create(m_DBConnection)};
    if (!SQL_SUCCEEDED(nRet))
    {
      LogODBC(m_DBConnection, _T("Failed to create ODBC statement while logging to database, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
      return false;
    }
  #pragma warning(suppress: 26473 26490)
    nRet = kODBCStatement.Prepare(reinterpret_cast<SQLTCHAR*>(sSQL.GetBuffer()));
    sSQL.ReleaseBuffer();
    if (!SQL_SUCCEEDED(nRet))
    {
      LogODBC(kODBCStatement, _T("Failed to prepare ODBC statement while logging to database, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
      return false;
    }

    //Bind the parameters
    const auto gd = record.m_LocalDateTime.GetDate();
    TIMESTAMP_STRUCT tsDateTime;
  #pragma warning(suppress: 26472)
    tsDateTime.year = static_cast<SQLSMALLINT>(gd.Year);
  #pragma warning(suppress: 26472)
    tsDateTime.month = static_cast<SQLUSMALLINT>(gd.Month);
  #pragma warning(suppress: 26472)
    tsDateTime.day = static_cast<SQLUSMALLINT>(gd.Day);
  #pragma warning(suppress: 26472)
    tsDateTime.hour = static_cast<SQLUSMALLINT>(gd.Hour);
  #pragma warning(suppress: 26472)
    tsDateTime.minute = static_cast<SQLUSMALLINT>(gd.Minute);
  #pragma warning(suppress: 26472)
    tsDateTime.second = static_cast<SQLUSMALLINT>(gd.Second);
    tsDateTime.fraction = 0;
    nRet = kODBCStatement.BindParameter(1, SQL_PARAM_INPUT, SQL_C_TIMESTAMP, SQL_TIMESTAMP, 23, 3, &tsDateTime, 0, nullptr);
    if (!SQL_SUCCEEDED(nRet))
    {
      LogODBC(kODBCStatement, _T("Failed to bind parameter 1 while logging to database, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
      return false;
    }
    SQLLEN nNull{SQL_NULL_DATA};
    if (record.m_nIndoorRelativeHumidity.has_value())
      nRet = kODBCStatement.BindParameter(2, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER, 0, 0, &record.m_nIndoorRelativeHumidity, 0, nullptr);
    else
      nRet = kODBCStatement.BindParameter(2, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER, 0, 0, nullptr, 0, &nNull);
    if (!SQL_SUCCEEDED(nRet))
    {
      LogODBC(kODBCStatement, _T("Failed to bind parameter 2 while logging to database, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
      return false;
    }
    if (record.m_fIndoorTemperature.has_value())
      nRet = kODBCStatement.BindParameter(3, SQL_PARAM_INPUT, SQL_C_FLOAT, SQL_FLOAT, 0, 0, &record.m_fIndoorTemperature, 0, nullptr);
    else
      nRet = kODBCStatement.BindParameter(3, SQL_PARAM_INPUT, SQL_C_FLOAT, SQL_FLOAT, 0, 0, nullptr, 0, &nNull);
    if (!SQL_SUCCEEDED(nRet))
    {
      LogODBC(kODBCStatement, _T("Failed to bind parameter 3 while logging to database, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
      return false;
    }
    if (record.m_nOutdoorRelativeHumidity.has_value())
      nRet = kODBCStatement.BindParameter(4, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER, 0, 0, &record.m_nOutdoorRelativeHumidity, 0, nullptr);
    else
      nRet = kODBCStatement.BindParameter(4, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER, 0, 0, nullptr, 0, &nNull);
    if (!SQL_SUCCEEDED(nRet))
    {
      LogODBC(kODBCStatement, _T("Failed to bind parameter 4 while logging to database, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
      return false;
    }
    if (record.m_fOutdoorTemperature.has_value())
      nRet = kODBCStatement.BindParameter(5, SQL_PARAM_INPUT, SQL_C_FLOAT, SQL_FLOAT, 0, 0, &record.m_fOutdoorTemperature, 0, nullptr);
    else
      nRet = kODBCStatement.BindParameter(5, SQL_PARAM_INPUT, SQL_C_FLOAT, SQL_FLOAT, 0, 0, nullptr, 0, &nNull);
    if (!SQL_SUCCEEDED(nRet))
    {
      LogODBC(kODBCStatement, _T("Failed to bind parameter 5 while logging to database, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
      return false;
    }
    if (record.m_fAbsolutePressure.has_value())
      nRet = kODBCStatement.BindParameter(6, SQL_PARAM_INPUT, SQL_C_FLOAT, SQL_FLOAT, 0, 0, &record.m_fAbsolutePressure, 0, nullptr);
    else
      nRet = kODBCStatement.BindParameter(6, SQL_PARAM_INPUT, SQL_C_FLOAT, SQL_FLOAT, 0, 0, nullptr, 0, &nNull);
    if (!SQL_SUCCEEDED(nRet))
    {
      LogODBC(kODBCStatement, _T("Failed to bind parameter 6 while logging to database, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
      return false;
    }
    if (record.m_fAverageWindSpeed.has_value())
      nRet = kODBCStatement.BindParameter(7, SQL_PARAM_INPUT, SQL_C_FLOAT, SQL_FLOAT, 0, 0, &record.m_fAverageWindSpeed, 0, nullptr);
    else
      nRet = kODBCStatement.BindParameter(7, SQL_PARAM_INPUT, SQL_C_FLOAT, SQL_FLOAT, 0, 0, nullptr, 0, &nNull);
    if (!SQL_SUCCEEDED(nRet))
    {
      LogODBC(kODBCStatement, _T("Failed to bind parameter 7 while logging to database, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
      return false;
    }
    if (record.m_fGustWindSpeed.has_value())
      nRet = kODBCStatement.BindParameter(8, SQL_PARAM_INPUT, SQL_C_FLOAT, SQL_FLOAT, 0, 0, &record.m_fGustWindSpeed, 0, nullptr);
    else
      nRet = kODBCStatement.BindParameter(8, SQL_PARAM_INPUT, SQL_C_FLOAT, SQL_FLOAT, 0, 0, nullptr, 0, &nNull);
    if (!SQL_SUCCEEDED(nRet))
    {
      LogODBC(kODBCStatement, _T("Failed to bind parameter 8 while logging to database, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
      return false;
    }
    if (record.m_nWindDirection.has_value())
      nRet = kODBCStatement.BindParameter(9, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER, 0, 0, &record.m_nWindDirection, 0, nullptr);
    else
      nRet = kODBCStatement.BindParameter(9, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER, 0, 0, nullptr, 0, &nNull);
    if (!SQL_SUCCEEDED(nRet))
    {
      LogODBC(kODBCStatement, _T("Failed to bind parameter 9 while logging to database, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
      return false;
    }
    float fTotalRain{0};
    if (record.m_nTotalRain.has_value())
    {
      fTotalRain = static_cast<float>(record.m_nTotalRain.value() * 0.3);
      nRet = kODBCStatement.BindParameter(10, SQL_PARAM_INPUT, SQL_C_FLOAT, SQL_FLOAT, 0, 0, &fTotalRain, 0, nullptr);
    }
    else
      nRet = kODBCStatement.BindParameter(10, SQL_PARAM_INPUT, SQL_C_FLOAT, SQL_FLOAT, 0, 0, nullptr, 0, &nNull);
    if (!SQL_SUCCEEDED(nRet))
    {
      LogODBC(kODBCStatement, _T("Failed to bind parameter 10 while logging to database, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
      return false;
    }
    if (record.m_fRainfallSinceMidnight.has_value())
      nRet = kODBCStatement.BindParameter(11, SQL_PARAM_INPUT, SQL_C_FLOAT, SQL_FLOAT, 0, 0, &record.m_fRainfallSinceMidnight, 0, nullptr);
    else
      nRet = kODBCStatement.BindParameter(11, SQL_PARAM_INPUT, SQL_C_FLOAT, SQL_FLOAT, 0, 0, nullptr, 0, &nNull);
    if (!SQL_SUCCEEDED(nRet))
    {
      LogODBC(kODBCStatement, _T("Failed to bind parameter 11 while logging to database, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
      return false;
    }
    if (record.m_fRainfall.has_value())
      nRet = kODBCStatement.BindParameter(12, SQL_PARAM_INPUT, SQL_C_FLOAT, SQL_FLOAT, 0, 0, &record.m_fRainfall, 0, nullptr);
    else
      nRet = kODBCStatement.BindParameter(12, SQL_PARAM_INPUT, SQL_C_FLOAT, SQL_FLOAT, 0, 0, nullptr, 0, &nNull);
    if (!SQL_SUCCEEDED(nRet))
    {
      LogODBC(kODBCStatement, _T("Failed to bind parameter 12 while logging to database, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
      return false;
    }
    if (record.m_nUV.has_value())
      nRet = kODBCStatement.BindParameter(13, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER, 0, 0, &record.m_nUV, 0, nullptr);
    else
      nRet = kODBCStatement.BindParameter(13, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER, 0, 0, nullptr, 0, &nNull);
    if (!SQL_SUCCEEDED(nRet))
    {
      LogODBC(kODBCStatement, _T("Failed to bind parameter 13 while logging to database, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
      return false;
    }
    if (record.m_fIlluminance.has_value())
      nRet = kODBCStatement.BindParameter(14, SQL_PARAM_INPUT, SQL_C_FLOAT, SQL_FLOAT, 0, 0, &record.m_fIlluminance, 0, nullptr);
    else
      nRet = kODBCStatement.BindParameter(14, SQL_PARAM_INPUT, SQL_C_FLOAT, SQL_FLOAT, 0, 0, nullptr, 0, &nNull);
    if (!SQL_SUCCEEDED(nRet))
    {
      LogODBC(kODBCStatement, _T("Failed to bind parameter 14 while logging to database, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
      return false;
    }

    //Execute the statement
    nRet = kODBCStatement.Execute();
    if (!SQL_SUCCEEDED(nRet))
    {
      //PJ TO DO: To do email notification on failure here

      LogODBC(kODBCStatement, _T("Failed to log record to database, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));

      //Throw away the connection if the insert failed prior to returning
      m_DBConnection.Close();
      return false;
    }
  }

  //Remember the time of this DB save
  m_LastDBSave = DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentSystemTime();

  //Remember the last rainfall count
  if (record.m_nTotalRain.has_value())
  {
    m_nLastRainfallCount = record.m_nTotalRain.value();
    m_LastRainfallCount = now;

    //Save the current rainfall value and datetime to allow us to reuse these values on service restart
    static const TCHAR* pszLastRainfallSection = _T("LastRainfall");
    WriteProfileInt(pszLastRainfallSection, _T("Count"), m_nLastRainfallCount);
    const auto gd = record.m_LocalDateTime.GetDate();
    WriteProfileInt(pszLastRainfallSection, _T("Year"), gd.Year);
    WriteProfileInt(pszLastRainfallSection, _T("Month"), gd.Month);
    WriteProfileInt(pszLastRainfallSection, _T("Day"), gd.Day);
    WriteProfileInt(pszLastRainfallSection, _T("Hour"), gd.Hour);
    WriteProfileInt(pszLastRainfallSection, _T("Minute"), gd.Minute);
    WriteProfileInt(pszLastRainfallSection, _T("Second"), gd.Second);

    if (m_bDebugAuditing)
      Log(_T("Remembering last rainfall count of %d for %s\n"), m_nLastRainfallCount, record.m_LocalDateTime.Format(_T("%x")).c_str());
  }

  return true;
}

void CMyService::OnStop() noexcept
{
  //Just signal the stop event
  m_StopEvent.Set();
}

void CMyService::OnShutdown() noexcept
{
  //Just signal the stop event
  m_StopEvent.Set();
}

bool CMyService::OpenDB(_In_ bool bAttemptSchemaCreation)
{
  //Check to see if we are using Sqlite3
  if (!m_sSqliteDatabase.IsEmpty())
  {
    //The Sqlite codepath

    //Nothing to do if already connected
    if (m_SqliteDB.operator sqlite3*())
      return true;

    Log(_T("Connecting to Sqlite database '%s'\n"), m_sSqliteDatabase.GetString());
    const int nOpen{m_SqliteDB.Open(ATL::CT2A{m_sSqliteDatabase, CP_UTF8})};
    if (nOpen != SQLITE_OK)
    {
      //PJ TO DO: To do email notification on failure here
      Log(_T("Failed to open Sqlite database '%s, Error:%d\n"), m_sSqliteDatabase.GetString(), nOpen);
      return false;
    }
    Log(_T("Connected to Sqlite database\n"));

    if (bAttemptSchemaCreation)
    {
      Log(_T("Creating 'historicalData' table in database if necessary\n"));

      //Form the SQL we will be using
      ATL::CStringA sSQL{" CREATE TABLE IF NOT EXISTS historicalData\r\n"
                         " (\r\n"
                         "  dateTime DATETIME PRIMARY KEY ASC NOT NULL,\r\n"
                         "  indoorHumidity INTEGER,\r\n"
                         "  indoorTemperature REAL,\r\n"
                         "  outdoorHumidity INTEGER,\r\n"
                         "  outdoorTemperature REAL,\r\n"
                         "  pressure REAL,\r\n"
                         "  avgWindSpeed REAL,\r\n"
                         "  gustWindSpeed REAL,\r\n"
                         "  windDirection INTEGER,\r\n"
                         "  totalRainfall REAL,\r\n"
                         "  rainfallSinceMidnight REAL,\r\n"
                         "  rainfall REAL,\r\n"
                         "  UV INTEGER,\r\n"
                         "  illuminance REAL\r\n"
                         " );"};
      char* pszErrorMessage{nullptr};
      const int nExec{m_SqliteDB.Exec(sSQL, nullptr, nullptr, &pszErrorMessage)};
      if ((nExec != SQLITE_OK) || (pszErrorMessage != nullptr))
      {
        Log(_T("Failed to create 'historicalData' database table, Error:%d, %s\n"), nExec, pszErrorMessage);
        sqlite3_free(pszErrorMessage);
        return false;
      }
      if (pszErrorMessage != nullptr)
      {
        sqlite3_free(pszErrorMessage);
        pszErrorMessage = nullptr;
      }
    }
  }
  else
  {
    //The ODBC codepath

    //Nothing to do if already connected
  #pragma warning(suppress: 26477)
    if (m_DBConnection != SQL_NULL_HANDLE)
      return true;

    //Setup the ODBC environment
    CODBC::CEnvironment env;
    Log(_T("Creating ODBC environment\n"));
    SQLRETURN nRet{env.Create()};
    if (!SQL_SUCCEEDED(nRet))
    {
      Log(_T("Failed to create ODBC environment, Error:%d\n"), static_cast<int>(nRet));
      return false;
    }

    //We want ODBC v3 behaviour
    nRet = env.SetAttr(SQL_ATTR_ODBC_VERSION, SQL_OV_ODBC3);
    if (!SQL_SUCCEEDED(nRet))
    {
      Log(_T("Failed to setup ODBC v3 behaviour, Error:%d\n"), static_cast<int>(nRet));
      return false;
    }

    //Create the ODBC connection
    Log(_T("Creating ODBC connection\n"));
    nRet = m_DBConnection.Create(env);
    if (!SQL_SUCCEEDED(nRet))
    {
      Log(_T("Failed to create ODBC connection, Error:%d\n"), static_cast<int>(nRet));
      return false;
    }

    //Connect to the ODBC connection
    Log(_T("Connecting to database, initialization string: %s\n"), m_sDBInitString.GetString());
    CODBC::String sFullInitString;
  #pragma warning(suppress: 26473 26490)
    nRet = m_DBConnection.DriverConnect(reinterpret_cast<SQLTCHAR*>(m_sDBInitString.GetBuffer()), sFullInitString);
    m_sDBInitString.ReleaseBuffer();
    if (!SQL_SUCCEEDED(nRet))
    {
      //PJ TO DO: To do email notification on failure here")
      LogODBC(m_DBConnection, _T("Failed to connect to database, initialization string:%s, Error:%d, <DBERROR>\n"), m_sDBInitString.GetString(), static_cast<int>(nRet));
      return false;
    }
    Log(_T("Connected to database\n"));

    if (bAttemptSchemaCreation)
    {
      Log(_T("Creating 'historicalData' table in database if necessary\n"));

      //Form the SQL we will be using
      ATL::CString sSQL{_T("IF NOT EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'[historicalData]') AND type in(N'U'))\r\n")
                        _T("BEGIN\r\n")
                        _T(" CREATE TABLE [historicalData]\r\n")
                        _T(" (\r\n")
                        _T("  [dateTime][datetime] NOT NULL,\r\n")
                        _T("  [indoorHumidity][int] NULL,\r\n")
                        _T("  [indoorTemperature][float] NULL,\r\n")
                        _T("  [outdoorHumidity][int] NULL,\r\n")
                        _T("  [outdoorTemperature][float] NULL,\r\n")
                        _T("  [pressure][float] NULL,\r\n")
                        _T("  [avgWindSpeed][float] NULL,\r\n")
                        _T("  [gustWindSpeed][float] NULL,\r\n")
                        _T("  [windDirection][int] NULL,\r\n")
                        _T("  [totalRainfall][float] NULL,\r\n")
                        _T("  [rainfallSinceMidnight][float] NULL,\r\n")
                        _T("  [rainfall][float] NULL,\r\n")
                        _T("  [UV][int] NULL,\r\n")
                        _T("  [illuminance][float] NULL,\r\n")
                        _T("  CONSTRAINT [PK_historicalData] PRIMARY KEY CLUSTERED\r\n")
                        _T("  (\r\n")
                        _T("   [dateTime] ASC\r\n")
                        _T("  )\r\n")
                        _T(" )\r\n")
                        _T("END\r\n")
                       };

      //Create and prepare the statement
      CODBC::CStatement kODBCStatement;
      nRet = kODBCStatement.Create(m_DBConnection);
      if (!SQL_SUCCEEDED(nRet))
      {
        LogODBC(m_DBConnection, _T("Failed to create ODBC statement while creating 'historicalData' database table, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
        return false;
      }
  #pragma warning(suppress: 26473 26490)
      nRet = kODBCStatement.Prepare(reinterpret_cast<SQLTCHAR*>(sSQL.GetBuffer()));
      sSQL.ReleaseBuffer();
      if (!SQL_SUCCEEDED(nRet))
      {
        LogODBC(kODBCStatement, _T("Failed to prepare ODBC statement while creating 'historicalData' database table, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
        return false;
      }

      //Execute the statement
      nRet = kODBCStatement.Execute();
      if (!SQL_SUCCEEDED(nRet))
      {
        LogODBC(kODBCStatement, _T("Failed to create 'historicalData' database table, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
        return false;
      }

      //create the upsert stored procedure
      Log(_T("Creating 'historicalDataUpsert' stored procedure in database if necessary\n"));

      sSQL = _T("IF NOT EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'[historicalDataUpsert]') AND type in(N'P', N'PC'))\r\n")
             _T("BEGIN\r\n")
             _T(" DECLARE @sql NVARCHAR(MAX)\r\n")
             _T(" SELECT @sql = 'CREATE PROCEDURE [historicalDataUpsert]\r\n")
             _T(" (\r\n")
             _T("  @dateTime datetime,\r\n")
             _T("  @indoorHumidity int,\r\n")
             _T("  @indoorTemperature float,\r\n")
             _T("  @outdoorHumidity int,\r\n")
             _T("  @outdoorTemperature float,\r\n")
             _T("  @pressure float,\r\n")
             _T("  @avgWindSpeed float,\r\n")
             _T("  @gustWindSpeed float,\r\n")
             _T("  @windDirection int,\r\n")
             _T("  @totalRainfall float,\r\n")
             _T("  @rainfallSinceMidnight float,\r\n")
             _T("  @rainfall float,\r\n")
             _T("  @UV int,\r\n")
             _T("  @illuminance float\r\n")
             _T(" )\r\n")
             _T(" AS\r\n")
             _T("  SET TRANSACTION ISOLATION LEVEL SERIALIZABLE\r\n")
             _T("  BEGIN TRAN\r\n")
             _T("  IF EXISTS (SELECT * FROM [historicalData] WITH(UPDLOCK) WHERE dateTime = @dateTime)\r\n")
             _T("   UPDATE [historicalData]\r\n")
             _T("   SET [indoorHumidity] = @indoorHumidity,\r\n")
             _T("       [indoorTemperature] = @indoorTemperature,\r\n")
             _T("       [outdoorHumidity] = @outdoorHumidity,\r\n")
             _T("       [outdoorTemperature] = @outdoorTemperature,\r\n")
             _T("       [pressure] = @pressure,\r\n")
             _T("       [avgWindSpeed] = @avgWindSpeed,\r\n")
             _T("       [gustWindSpeed] = @gustWindSpeed,\r\n")
             _T("       [windDirection] = @windDirection,\r\n")
             _T("       [totalRainfall] = @totalRainfall,\r\n")
             _T("       [rainfallSinceMidnight] = @rainfallSinceMidnight,\r\n")
             _T("       [rainfall] = @rainfall,\r\n")
             _T("       [UV] = @UV,\r\n")
             _T("       [illuminance] = @illuminance\r\n")
             _T("   WHERE dateTime = @DateTime\r\n")
             _T("  ELSE\r\n")
             _T("   INSERT INTO [historicalData] ([dateTime], [indoorHumidity], [indoorTemperature], [outdoorHumidity], [outdoorTemperature], [pressure], [avgWindSpeed], [gustWindSpeed], [windDirection], [totalRainfall], [rainfallSinceMidnight], [rainfall], [UV], [illuminance])\r\n")
             _T("   VALUES (@dateTime, @indoorHumidity, @indoorTemperature, @outdoorHumidity, @outdoorTemperature, @pressure, @avgWindSpeed, @gustWindSpeed, @windDirection, @totalRainfall, @rainfallSinceMidnight, @rainfall, @UV, @illuminance)\r\n")
             _T("  COMMIT'\r\n")
             _T(" EXEC sp_executesql @SQL\r\n")
             _T("END\r\n");

      //Create and prepare the statement
      kODBCStatement.Close();
      nRet = kODBCStatement.Create(m_DBConnection);
      if (!SQL_SUCCEEDED(nRet))
      {
        LogODBC(m_DBConnection, _T("Failed to create ODBC statement while creating 'historicalDataUpsert' stored procedure, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
        return false;
      }
  #pragma warning(suppress: 26473 26490)
      nRet = kODBCStatement.Prepare(reinterpret_cast<SQLTCHAR*>(sSQL.GetBuffer()));
      sSQL.ReleaseBuffer();
      if (!SQL_SUCCEEDED(nRet))
      {
        LogODBC(kODBCStatement, _T("Failed to prepare ODBC statement while creating 'historicalDataUpsert' stored procedure, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
        return false;
      }

      //Execute the statement
      nRet = kODBCStatement.Execute();
      if (!SQL_SUCCEEDED(nRet))
      {
        LogODBC(kODBCStatement, _T("Failed to create 'historicalDataUpsert' stored procedure, Error:%d, <DBERROR>\n"), static_cast<int>(nRet));
        return false;
      }
    }
  }

  return true;
}

bool CMyService::LogWin32Error(_In_ DWORD dwLastError, _In_z_ const TCHAR* pszFormat, ...)
{
  ATL::CString sLastError;
  sLastError.Format(_T("%u"), dwLastError);

#pragma warning(suppress: 26826)
  va_list argp;
#pragma warning(suppress: 26492 26481 26826)
  va_start(argp, pszFormat);
  ATL::CString s;
  s.FormatV(pszFormat, argp);
#pragma warning(suppress: 26477 26826)
  va_end(argp);
  if (dwLastError == ERROR_TIMEOUT)
  {
    if (m_bResetUsingYKUSHXS)
      s.Replace(_T("<WIN32ERROR>"), sLastError + _T(", this error was due to a timeout on the device when reading or writing to it. Most likely the USB port on the device has locked up. Will attempt reset using YKUSH XS."));
    else
      s.Replace(_T("<WIN32ERROR>"), sLastError + _T(", this error was due to a timeout on the device when reading or writing to it. Most likely the USB port on the device has locked up and the device will need to be hard reset."));
  }
  else
    s.Replace(_T("<WIN32ERROR>"), sLastError);

  //Delegate to the other version of the method
  return Log(_T("%s"), s.GetString());
}

bool CMyService::LogODBC(_In_ const CODBC::CHandle& handle, _In_z_ const TCHAR* pszFormat, ...)
{
  std::vector<CODBC::String> records;
  handle.GetDiagRecords(records);
  ATL::CString sRecords;
  for (size_t i{0}; i<records.size(); i++)
  {
#pragma warning(suppress: 26446)
    sRecords += records[i].c_str();
    if (i < (records.size() - 1))
      sRecords += _T(", ");
  }

#pragma warning(suppress: 26826)
  va_list argp;
#pragma warning(suppress: 26492 26481 26826)
  va_start(argp, pszFormat);
  ATL::CString s;
  s.FormatV(pszFormat, argp);
#pragma warning(suppress: 26477 26826)
  va_end(argp);
  s.Replace(_T("<DBERROR>"), sRecords);

  //Delegate to the other version of the method
  return Log(_T("%s"), s.GetString());
}

bool CMyService::Log(_In_z_ const TCHAR* pszFormat, ...)
{
  //Write to the console and to the disk log
#pragma warning(suppress: 26826)
  va_list argp;
#pragma warning(suppress: 26492 26481 26826)
  va_start(argp, pszFormat);
  ATL::CString s;
  s.FormatV(pszFormat, argp);
#pragma warning(suppress: 26477 26826)
  va_end(argp);
  _tprintf(_T("%s"), s.GetString());

  //Reopen the log file if necessary
  const auto now{DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentLocalTime()};
  ATL::CString sNewLogFile{m_sLogDirectory};
  sNewLogFile += _T("\\");
  tm tmTemp{};
  if (!now.GetAsTM(tmTemp))
    return false;
  ATL::CString sFile;
  LPTSTR pszFile = sFile.GetBufferSetLength(4096);
  _tcsftime(pszFile, sFile.GetLength(), m_sLogFilenameSpecifier, &tmTemp);
  sFile.ReleaseBuffer();
  sNewLogFile += sFile;
  if (sNewLogFile != m_sCurrentLogFilename)
  {
    if (m_Log.operator HANDLE() != nullptr)
      m_Log.Close();
    const HRESULT hr{m_Log.Create(sNewLogFile, GENERIC_WRITE, FILE_SHARE_READ, OPEN_ALWAYS)};
    if (FAILED(hr))
      return false;
    m_Log.Seek(0, FILE_END);
    m_sCurrentLogFilename = sNewLogFile;
  }
  if (m_Log.operator HANDLE() == nullptr)
    return false;

  //Include the date / time on each log line
  ATL::CString s2(now.Format(_T("%x")).c_str());
  s = s2 + _T(" ") + s;

  //Note we write the log file out as UTF8
#ifdef _UNICODE
  ATL::CW2A sUTF8{s, CP_UTF8};
#else
  ATL::CW2A sUTF8{ATL::CA2W{s}, CP_UTF8};
#endif
#pragma warning(suppress: 26472)
  const HRESULT hr{m_Log.Write(sUTF8.operator LPSTR(), static_cast<DWORD>(strlen(sUTF8)))};
  return SUCCEEDED(hr);
}

bool CMyService::SynchronizeWithStation()
{
  if ((m_StationClock.IsValid()) && (m_SensorClock.IsValid()) && !SafeToReadFromDevice())
  {
    Log(_T("Not synchronizing with weather station because not safe to do so at this time\n"));
    return false;
  }

  //Nothing to do if the device is currently not open
  if (m_Device.operator HANDLE() == INVALID_HANDLE_VALUE)
    return false;

#ifdef FOSTATION_UPDATE_READ_PERIOD
  //Some test code to set the read period to 1 minute on the device
  ATLVERIFY(m_Device.WriteAddress(16, 0x1, m_dwDeviceTimeout)); //lets use a 1 minute read period
  ATLVERIFY(m_Device.WriteAddress(26, 0xAA, m_dwDeviceTimeout));
  std::vector<BYTE> byUpdate;
  while (m_Device.ReadAddress(26, byUpdate, m_dwDeviceTimeout) && (byUpdate[0] != 0x00));
#endif //#ifdef _DEBUG

  std::vector<BYTE> data;
  if (!m_Device.ReadAddress(0x0, data, m_dwDeviceTimeout))
  {
    //PJ TO DO: To do email notification on failure to read here
    const DWORD dwLastError{GetLastError()};
    LogWin32Error(dwLastError, _T("An error occurred reading address 0x0, Error:<WIN32ERROR>\n"));
    if (m_bResetUsingYKUSHXS && (dwLastError == ERROR_TIMEOUT))
      ResetUSBWithYKUSHXS();

    //Close down the weather station handle
    Log(_T("Closing weather station device\n"));
    m_Device.Close();

    return false;
  }
  Log(_T("Synchronizing with weather station (this may take a number of minutes depending on the current read interval)\n"));
#pragma warning(suppress: 26446)
  Log(_T("Current read interval: %d minute(s)\n"), static_cast<int>(data[16]));

  int nCurrentWriteAddress{-1};
  int nPreviousWriteAddress{-1};
  std::vector<BYTE> prevData;
  m_StationClock.Set(); //Mark "m_StationClock" as invalid
  m_SensorClock.Set(); //Mark "m_SensorClock" as invalid
  bool bStop{false};
  while (!bStop)
  {
#pragma warning(suppress: 26446)
    nCurrentWriteAddress = (data[CFOStation::CURRENT_LOCATION_ADDRESSMSB] * 256) + data[CFOStation::CURRENT_LOCATION_ADDRESSLSB];

    //Check to see if we need to exit prematurely from the loop
    bStop = WaitForSingleObject(m_StopEvent, 0) == WAIT_OBJECT_0;
    if (!bStop)
    {
      if (!m_SensorClock.IsValid())
      {
        //If the current write location has not changed, read the contents and see if it has changed
        if (nCurrentWriteAddress == nPreviousWriteAddress)
        {
          std::vector<BYTE> currentData;
          if (!m_Device.ReadAddress(nCurrentWriteAddress, currentData, m_dwDeviceTimeout))
          {
            //PJ TO DO: To do email notification on failure to read here
            const DWORD dwLastError{GetLastError()};
            LogWin32Error(dwLastError, _T("An error occurred reading address 0x%04X, Error:<WIN32ERROR>\n"), nCurrentWriteAddress);
            if (m_bResetUsingYKUSHXS && (dwLastError == ERROR_TIMEOUT))
              ResetUSBWithYKUSHXS();

            //Close down the weather station handle
            Log(_T("Closing weather station device\n"));
            m_Device.Close();

            return false;
          }
          currentData.resize(CFORecord::WITHOUT_SOLAR_RECORDSIZE); //We are only interested in the "without solar" record data
          if (prevData.size() && (currentData != prevData))
          {
            m_SensorClock = DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentLocalTime();
            Log(_T("Contents of current write address of 0x%04X has changed, determined sensor clock time to be %s\n"), nCurrentWriteAddress, m_SensorClock.Format(_T("%x")).c_str());
          }
          prevData = currentData;
        }
      }
      else
      {
        //See if the current write location has changed
        if ((nPreviousWriteAddress != -1) && (nCurrentWriteAddress != nPreviousWriteAddress))
        {
          m_StationClock = DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentLocalTime();
          Log(_T("Current write address has changed from 0x%04X to 0x%04X, determined station clock time to be %s\n"), nPreviousWriteAddress, nCurrentWriteAddress, m_StationClock.Format(_T("%x")).c_str());
        }
      }

      //Prepare for the next time around the loop
      nPreviousWriteAddress = nCurrentWriteAddress;
      bStop = (m_StationClock.IsValid() && m_SensorClock.IsValid());
      if (!bStop)
      {
        Sleep(500);

        //Get the address of the current location from the device
        if (!m_Device.ReadAddress(0x0, data, m_dwDeviceTimeout))
        {
          //PJ TO DO: To do email notification on failure to read here
          const DWORD dwLastError{GetLastError()};
          LogWin32Error(dwLastError, _T("An error occurred reading address 0x0, Error:<WIN32ERROR>\n"));
          if (m_bResetUsingYKUSHXS && (dwLastError == ERROR_TIMEOUT))
            ResetUSBWithYKUSHXS();

          //Close down the weather station handle
          Log(_T("Closing weather station device\n"));
          m_Device.Close();

          return false;
        }
      }
    }
  }
  return (WaitForSingleObject(m_StopEvent, 0) != WAIT_OBJECT_0);
}

bool CMyService::ResetUSBWithYKUSHXS()
{
  //PJ TO DO: To do email notification on YKUSH USB reset here

  Log(_T("Powering down USB port using YKUSH XS\n"));
  if (!m_YKUSHFXDevice.PortDown())
  {
    Log(_T("An error occurred powering down the USB port, Error:%u\n"), GetLastError());
    return false;
  }

  //Wait a while before we power up the port again
  const bool bStop{WaitForSingleObject(m_StopEvent, 1000) == WAIT_OBJECT_0};
  if (bStop)
    return false;

  Log(_T("Powering up USB port using YKUSH XS\n"));
  if (!m_YKUSHFXDevice.PortUp())
  {
    Log(_T("An error occurred powering up the USB port, Error:%u\n"), GetLastError());
    return false;
  }

  return true;
}

bool CMyService::ReadHistoricalData()
{
  Log(_T("Determining historical record data available\n"));
  //Get the address of the current location from the device
  std::vector<BYTE> data;
  bool bSuccess{m_Device.ReadAddress(0x0, data, m_dwDeviceTimeout)};
  if (!bSuccess)
  {
    LogWin32Error(GetLastError(), _T("An error occurred reading address 0x0, Error:<WIN32ERROR>\n"));
    return false;
  }
#pragma warning(suppress: 26446)
  int nCurrentAddress{(data[CFOStation::CURRENT_LOCATION_ADDRESSMSB] * 256) + data[CFOStation::CURRENT_LOCATION_ADDRESSLSB]};
  Log(_T("Current location address: 0x%04X\n"), nCurrentAddress);

  const auto localDateTime{DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentLocalTime()};
  const auto gd = localDateTime.GetDate();
  DTimePlus::CLDate<DTimePlus::CGregorianCalendar> currentLocal{gd.Year, gd.Month, gd.Day, gd.Hour, gd.Minute, 0};
  const auto gd2 = m_StationClock.GetDate();
  currentLocal += DTimePlus::CLTimeSpan{0, 0, 1, gd2.Second};
  Log(_T("Determined next expected write time will be %s\n"), currentLocal.Format(_T("%x")).c_str());
  const auto systemDateTime{DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentSystemTime()};
  const auto gd3 = systemDateTime.GetDate();
  DTimePlus::CLDate<DTimePlus::CGregorianCalendar> currentSystem{gd3.Year, gd3.Month, gd3.Day, gd3.Hour, gd3.Minute, 0};
  currentSystem += DTimePlus::CLTimeSpan{0, 0, 1, gd2.Second};

#pragma warning(suppress: 26446)
  const int nHistoricalRecords{(data[CFOStation::HISTORICAL_RECORDS_ADDRESSMSB] * 256) + data[CFOStation::HISTORICAL_RECORDS_ADDRESSLSB]};
  Log(_T("Historical records available: %d\n"), nHistoricalRecords);
  const int nMaxHistoricalRecords{min(nHistoricalRecords, static_cast<int>(m_nMaxHistoricalRecords))};
  Log(_T("Historical records to be read: %d\n"), nMaxHistoricalRecords);

  //Loop around reading all the historical data records
  std::vector<CFORecord> historicalRecords;
  historicalRecords.reserve(nMaxHistoricalRecords);
  bool bStop{nMaxHistoricalRecords == 0};
  int nRecordsProcessed{0};
  while (!bStop)
  {
    //Decrement the current address
    switch (m_StationType)
    {
      case FOStationType::WithSolarData:
      {
        nCurrentAddress -= CFORecord::WITH_SOLAR_RECORDSIZE;
        if (nCurrentAddress < CFOStation::MIN_LOCATION_ADDRESS)
        {
          nCurrentAddress = CFOStation::MAX_SOLAR_LOCATION_ADDRESS;
          Log(_T("Wrapping current address to 0x%04X\n"), nCurrentAddress);
        }
        break;
      }
      case FOStationType::WithourSolarData:
      {
        nCurrentAddress -= CFORecord::WITHOUT_SOLAR_RECORDSIZE;
        if (nCurrentAddress < CFOStation::MIN_LOCATION_ADDRESS)
        {
          nCurrentAddress = CFOStation::MAX_WITHOUTSOLAR_LOCATION_ADDRESS;
          Log(_T("Wrapping current address to 0x%04X\n"), nCurrentAddress);
        }
        break;
      }
      default:
      {
#pragma warning(suppress: 26477)
        ATLASSERT(FALSE);
        break;
      }
    }

    //Read the historical record if safe to do so at this time
    if (!SafeToReadFromDevice())
    {
      if (m_bDebugAuditing)
        Log(_T("Skipping read because not safe to do so at this time\n"));
      bStop = WaitForSingleObject(m_StopEvent, 500) == WAIT_OBJECT_0;
    }
    else
    {
      Log(_T("Reading historical data %d of %d at address 0x%04X\n"), nRecordsProcessed + 1, nMaxHistoricalRecords, nCurrentAddress);
      bSuccess = m_Device.ReadAddress(nCurrentAddress, data, m_dwDeviceTimeout);
      if (!bSuccess)
      {
        LogWin32Error(GetLastError(), _T("An error occurred reading historical data at address 0x%04X, Error:<WIN32ERROR>\n"), nCurrentAddress);
        return false;
      }
      switch (m_StationType)
      {
        case FOStationType::WithSolarData:
        {
          if (data.size() > CFORecord::WITH_SOLAR_RECORDSIZE)
            data.resize(CFORecord::WITH_SOLAR_RECORDSIZE);
          break;
        }
        case FOStationType::WithourSolarData:
        {
          if (data.size() > CFORecord::WITHOUT_SOLAR_RECORDSIZE)
            data.resize(CFORecord::WITHOUT_SOLAR_RECORDSIZE);
          break;
        }
        default:
        {
  #pragma warning(suppress: 26477)
          ATLASSERT(FALSE);
          break;
        }
      }

      //Decrement the calculated times for the historical record we have just read
#pragma warning(suppress: 26446)
      const auto ts{DTimePlus::CLTimeSpan{0, 0, data[0], 0}};
      currentLocal -= ts;
      currentSystem -= ts;

      //parse and process the record
      CFORecord record;
      if (record.ParseFromUSBHID(data))
      {
        record.m_LocalDateTime = currentLocal;
        record.m_UTCDateTime = currentSystem;

#pragma warning(suppress: 26486)
        historicalRecords.insert(historicalRecords.begin(), record);
      }
      else
        Log(_T("Could not parse historical record at address 0x%04X\n"), nCurrentAddress);
      ++nRecordsProcessed;
    }

    //Prepare for the next time around
    if (!bStop)
      bStop = WaitForSingleObject(m_StopEvent, 0) == WAIT_OBJECT_0;
    if (!bStop)
      bStop = (nRecordsProcessed >= nMaxHistoricalRecords);
  }
  Log(_T("Completed reading historical data\n"));

  if (WaitForSingleObject(m_StopEvent, 0) == WAIT_OBJECT_0)
    return false;

  Log(_T("Historical records to be stored: %d\n"), nMaxHistoricalRecords);
  bStop = (historicalRecords.size() == 0);
  nRecordsProcessed = 0;
  while (!bStop)
  {
#pragma warning(suppress: 26446)
    CFORecord& record{historicalRecords[nRecordsProcessed]};
#pragma warning(suppress: 26486)
    ProcessHistoricalRecord(record);
    ++nRecordsProcessed;

    //Prepare for the next time around
    if (!bStop)
      bStop = WaitForSingleObject(m_StopEvent, 0) == WAIT_OBJECT_0;
    if (!bStop)
      bStop = (nRecordsProcessed >= nMaxHistoricalRecords);
  }
  Log(_T("Completed storing historical data\n"));

  return (WaitForSingleObject(m_StopEvent, 0) != WAIT_OBJECT_0);
}

bool CMyService::SafeToReadFromDevice() noexcept
{
  if (m_Device.operator HANDLE() == INVALID_HANDLE_VALUE)
    return false;

  //Validate our parameters
  ATLASSERT(m_StationClock.IsValid());
  ATLASSERT(m_SensorClock.IsValid());

  const DTimePlus::CLDate<DTimePlus::CGregorianCalendar> now{DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentLocalTime()};
  auto ts{now - m_StationClock};
  int nTotalSeconds{static_cast<int>(ts.SecondsAsDouble())};
  const bool bStationClockOk{((nTotalSeconds % 60) > m_nAvoidReadPeriod) && ((nTotalSeconds % 60) < (60 - m_nAvoidReadPeriod))};
  ts = now - m_SensorClock;
  nTotalSeconds = static_cast<int>(ts.SecondsAsDouble());
  const bool bSensorClockOk{((nTotalSeconds % 48) > m_nAvoidReadPeriod) && ((nTotalSeconds % 48) < (48 - m_nAvoidReadPeriod))};
  return bStationClockOk && bSensorClockOk;
}

bool CMyService::OpenWeatherStationAsHIDDevice()
{
  //Open up the HID Fine Offset device (in overlapped mode if required to allow us to implement timed out read calls)
#pragma warning(suppress: 26446)
  Log(_T("Opening weather station device '%s'\n"), m_sDevicePath.c_str());
#pragma warning(suppress: 26446)
  const bool bSuccess{m_Device.Open(m_sDevicePath.c_str(), GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, nullptr, m_dwDeviceTimeout ? FILE_FLAG_OVERLAPPED : 0) != FALSE};
  if (!bSuccess)
  {
    m_Device.Close();
    //PJ TO DO: To do email notification on failure here
  }

  return bSuccess;
}

void CMyService::ReadCOMPortData(_Inout_ ATL::CStringA& sBuffer, _Inout_ OVERLAPPED& overlapped, _Inout_ CFORecord& lastWeatherRecord, _Inout_ CFORecord& lastLightRecord)
{
  //Try to read something from the port in overlapped mode
  DWORD dwBytesRead{0};
  std::array<char, 1024> buffer{};
#pragma warning(suppress: 26472 26485)
  if (ReadFile(m_Port, buffer.data(), static_cast<DWORD>(buffer.size() - 1), &dwBytesRead, &overlapped) && dwBytesRead)
  {
    //Null terminate the received data
#pragma warning(suppress: 26446 26482)
    buffer[dwBytesRead] = '\0';

    //Process the received data
#pragma warning(suppress: 26485)
    HandleCOMPortData(buffer.data(), sBuffer, lastWeatherRecord, lastLightRecord);
  }
}

void CMyService::HandleCOMPortData(_In_z_ LPCSTR pszNewData, _Inout_ ATL::CStringA& sBuffer, _Inout_ CFORecord& lastWeatherRecord, _Inout_ CFORecord& lastLightRecord)
{
  //Append the newly read data to the buffer
  sBuffer += pszNewData;

  //See if we've got a sentence yet
  int nLocationDelimiter{sBuffer.Find('\n')};
  if (nLocationDelimiter == -1)
    nLocationDelimiter = sBuffer.Find('\r');

  while (nLocationDelimiter != -1)
  {
    //Yes, there is a line feed in the string
    ATL::CStringA sLine{sBuffer.Left(nLocationDelimiter - 1)};

    //Now trim off the sentence from what came in from the port
    sBuffer = sBuffer.Right(sBuffer.GetLength() - (nLocationDelimiter + 1));
    nLocationDelimiter = sBuffer.Find('\n');
    if (nLocationDelimiter == -1)
      nLocationDelimiter = sBuffer.Find('\r');

    //Log the raw data received if debug auditing is on
    if (m_bDebugAuditing)
      Log(_T("Received: %s\n"), ATL::CA2T{sLine}.operator LPTSTR());

    //Interpret the receive line data
    if (sLine.Find("MSG:Weather") == 0)
    {
      //Parse the weather message and timestamp it
      lastWeatherRecord.ParseFromCOMWeatherMsg(sLine);
      lastWeatherRecord.m_LocalDateTime = DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentLocalTime();
      lastWeatherRecord.m_UTCDateTime = DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentSystemTime();

      if (m_bDebugAuditing)
        LogRawRecord(_T("Weather Record: %s\n"), lastWeatherRecord);

      if (m_StationType == FOStationType::WithSolarData)
      {
        //Copy the values only available in light messages from the last light message if they are within the last few minutes
        if (lastLightRecord.m_UTCDateTime.IsValid())
        {
          if ((lastWeatherRecord.m_UTCDateTime - lastLightRecord.m_UTCDateTime) < DTimePlus::CLTimeSpan(0, 0, 3, 0))
          {
            if (m_bDebugAuditing)
              Log(_T("Copying last light message\n"));
            lastWeatherRecord.m_nSensorID = lastLightRecord.m_nSensorID;
            lastWeatherRecord.m_bLightStatusOK = lastLightRecord.m_bLightStatusOK;
            lastWeatherRecord.m_nUV = lastLightRecord.m_nUV;
            lastWeatherRecord.m_fIlluminance = lastLightRecord.m_fIlluminance;
          }
          else
          {
            //PJ TO DO: To do email notification on no light messages in 3 minutes
            Log(_T("Not copying last light message as too old\n"));
          }
        }
        else
          Log(_T("Last light message for %s is not valid\n"), lastWeatherRecord.m_LocalDateTime.Format(_T("%x")).c_str());
      }

      if (m_bDebugAuditing)
        Log(_T("Processing record for %s\n"), lastWeatherRecord.m_LocalDateTime.Format(_T("%x")).c_str());
      ProcessCurrentRecord(lastWeatherRecord);

      //Now reset the last weather record now that we have processed it
      lastWeatherRecord = CFORecord();
    }
    else if (sLine.Find("MSG:Light") == 0)
    {
      //Parse the light message and timestamp it
      CFORecord tempLightRecord;
      tempLightRecord.ParseFromCOMLightMsg(sLine);
      tempLightRecord.m_LocalDateTime = DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentLocalTime();
      tempLightRecord.m_UTCDateTime = DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentSystemTime();

      if (m_bDebugAuditing)
        LogRawLightRecord(_T("Light Record: %s\n"), tempLightRecord);

      //Only use the last parsed light message if its LightStatus value is ok
      if (tempLightRecord.m_bLightStatusOK.has_value())
      {
        if (tempLightRecord.m_bLightStatusOK.value())
          lastLightRecord = tempLightRecord;
        else
          Log(_T("Not using last light message as its LightStatus value is not OK\n"));
      }
      else
        lastLightRecord = tempLightRecord;
    }
  }
}

void CMyService::ServiceMain(_In_ DWORD /*dwArgc*/, _In_opt_ LPTSTR* /*pszArgv*/)
{
  ReportStatus(SERVICE_START_PENDING, 1, 1000);

  //Read in our configuration values
  static const TCHAR* pszGeneralSection{_T("General")};
  m_StationType = static_cast<FOStationType>(GetProfileInt(pszGeneralSection, _T("StationType"), static_cast<int>(m_StationType)));
  m_sLogFilenameSpecifier = GetProfileString(pszGeneralSection, _T("LogFilenameSpecifier"), m_sLogFilenameSpecifier).c_str();
  m_sDBInitString = GetProfileString(pszGeneralSection, _T("DBInitString"), m_sDBInitString).c_str();
  m_sSqliteDatabase = GetProfileString(pszGeneralSection, _T("SqliteDatabase"), m_sSqliteDatabase).c_str();
  m_nAvoidReadPeriod = GetProfileInt(pszGeneralSection, _T("AvoidReadPeriod"), m_nAvoidReadPeriod);
  m_nMaxHistoricalRecords = GetProfileInt(pszGeneralSection, _T("MaxHistoricalRecords"), m_nMaxHistoricalRecords);
  m_dwDeviceTimeout = GetProfileInt(pszGeneralSection, _T("DeviceTimeout"), m_dwDeviceTimeout);
  m_sMETWOWSiteID = GetProfileString(pszGeneralSection, _T("METWOWSiteID"), m_sMETWOWSiteID).c_str();
  m_sMETWOWSiteAuthenticationKey = GetProfileString(pszGeneralSection, _T("METWOWSiteAuthenticationKey"), m_sMETWOWSiteAuthenticationKey).c_str();
  m_sWeatherUndergroundID = GetProfileString(pszGeneralSection, _T("WeatherUndergroundID"), m_sWeatherUndergroundID).c_str();
  m_sWeatherUndergroundPassword = GetProfileString(pszGeneralSection, _T("WeatherUndergroundPassword"), m_sWeatherUndergroundPassword).c_str();
#pragma warning(suppress: 26472)
  m_bDebugAuditing = static_cast<bool>(GetProfileInt(pszGeneralSection, _T("DebugAuditing"), m_bDebugAuditing));
#pragma warning(suppress: 26472)
  m_bResetUsingYKUSHXS = static_cast<bool>(GetProfileInt(pszGeneralSection, _T("ResetUsingYKUSHXS"), m_bResetUsingYKUSHXS));
  m_nMETWOWLoggingInterval = GetProfileInt(pszGeneralSection, _T("METWOWLoggingInterval"), m_nMETWOWLoggingInterval);
  m_nWeatherUndergroundLoggingInterval = GetProfileInt(pszGeneralSection, _T("WeatherUndergroundLoggingInterval"), m_nWeatherUndergroundLoggingInterval);
  m_nDBLoggingInterval = GetProfileInt(pszGeneralSection, _T("DBLoggingInterval"), m_nDBLoggingInterval);
  m_nCOMPort = GetProfileInt(pszGeneralSection, _T("COMPort"), m_nCOMPort);
  m_nBaudRate = GetProfileInt(pszGeneralSection, _T("BaudRate"), m_nBaudRate);

  //Cache the version info details of this exe
  CVersionInfo info;
  ATL::CString sApp;
  GetModuleFileName(nullptr, sApp.GetBuffer(_MAX_PATH), _MAX_PATH);
  sApp.ReleaseBuffer();
  if (info.Load(sApp))
  {
    const VS_FIXEDFILEINFO* pffi{info.GetFixedFileInfo()};
    m_sVersionInfo.Format(_T("%d.%d.%d.%d"), static_cast<int>(HIWORD(pffi->dwProductVersionMS)), static_cast<int>(LOWORD(pffi->dwProductVersionMS)),
                          static_cast<int>(HIWORD(pffi->dwProductVersionLS)), static_cast<int>(LOWORD(pffi->dwProductVersionLS)));
  }

  //Lets use a log file in the "FOLDERID_LocalAppData"\FOWeather directory
  PWSTR pwszCacheDirectory{nullptr};
#pragma warning(suppress: 26496)
  HRESULT hr{SHGetKnownFolderPath(FOLDERID_LocalAppData, KF_FLAG_DEFAULT, nullptr, &pwszCacheDirectory)};
  if (FAILED(hr))
    TerminateService(hr);
  m_sLogDirectory = pwszCacheDirectory;
  CoTaskMemFree(pwszCacheDirectory);
  m_sLogDirectory += _T("\\FOStation");
  CreateDirectory(m_sLogDirectory, nullptr);
  Log(_T("******* Service starting *******\n"));

  //Log our current config settings
  Log(_T("Station type: %d\n"), static_cast<int>(m_StationType));
  Log(_T("LogFilenameSpecifier: %s\n"), m_sLogFilenameSpecifier.GetString());
  Log(_T("DBInitString: %s\n"), m_sDBInitString.GetString());
  Log(_T("SqliteDatabase: %s\n"), m_sSqliteDatabase.GetString());
  Log(_T("AvoidReadPeriod (seconds): %d\n"), m_nAvoidReadPeriod);
  Log(_T("MaxHistoricalRecords: %d\n"), m_nMaxHistoricalRecords);
  Log(_T("DeviceTimeout (milliseconds): %u\n"), m_dwDeviceTimeout);
  Log(_T("DebugAuditing: %d\n"), static_cast<int>(m_bDebugAuditing));
  Log(_T("ResetUsingYKUSHXS: %d\n"), static_cast<int>(m_bResetUsingYKUSHXS));
  Log(_T("MET WOW Site ID: %s\n"), m_sMETWOWSiteID.GetString());
  if (!m_sMETWOWSiteAuthenticationKey.IsEmpty())
    Log(_T("MET WOW Site authentication key: ******\n"));
  else
    Log(_T("MET WOW Site authentication key:\n"));
  Log(_T("MET WOW logging interval: %d\n"), m_nMETWOWLoggingInterval);
  Log(_T("Weather Underground ID: %s\n"), m_sWeatherUndergroundID.GetString());
  if (!m_sWeatherUndergroundPassword.IsEmpty())
    Log(_T("Weather Underground password: ******\n"));
  else
    Log(_T("Weather Underground password:\n"));
  Log(_T("Weather Underground logging interval: %d\n"), m_nWeatherUndergroundLoggingInterval);
  Log(_T("DB Logging interval: %d\n"), m_nDBLoggingInterval);
  Log(_T("Product Version info: %s\n"), m_sVersionInfo.GetString());
  Log(_T("COM Port: %d\n"), m_nCOMPort);
  Log(_T("BaudRate: %d\n"), m_nBaudRate);

  //check to see if we can use the cached midnight rainfall value
  static const TCHAR* pszMidnightRainfallSection{_T("MidnightRainfall")};
#pragma warning(suppress: 26472)
  const int nMidnightRainfallCount{static_cast<int>(GetProfileInt(pszMidnightRainfallSection, _T("Count"), INT_MIN))};
#pragma warning(suppress: 26472)
  const int nYearMidnightRainfall{static_cast<int>(GetProfileInt(pszMidnightRainfallSection, _T("Year"), 0))};
#pragma warning(suppress: 26472)
  const int nMonthMidnightRainfall{static_cast<int>(GetProfileInt(pszMidnightRainfallSection, _T("Month"), 0))};
#pragma warning(suppress: 26472)
  const int nDayMidnightRainfall{static_cast<int>(GetProfileInt(pszMidnightRainfallSection, _T("Day"), 0))};
  auto now{DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentLocalTime()};
  const auto gd{now.GetDate()};
  const int nNewYear{gd.Year};
  const int nNewMonth{gd.Month};
  const int nNewDay{gd.Day};
  if ((nMidnightRainfallCount != INT_MIN) && (nNewYear == nYearMidnightRainfall) &&
      (nNewMonth == nMonthMidnightRainfall) && (nNewDay == nDayMidnightRainfall))
  {
    m_nMidnightRainfallCount = nMidnightRainfallCount;
    m_nYearMidnightRainfall = nNewYear;
    m_nMonthMidnightRainfall = nNewMonth;
    m_nDayMidnightRainfall = nNewDay;
    Log(_T("Reloading rainfall count of %d for today\n"), m_nMidnightRainfallCount);
  }
  else
  {
    Log(_T("Not Reloading rainfall count of %d for today as persisted value is not for today or is invalid, Year=%d, Month=%d, Day=%d\n"), m_nMidnightRainfallCount, nYearMidnightRainfall, nMonthMidnightRainfall, nDayMidnightRainfall);
  }

  //Check to see if we can use the cached last rainfall value
  static const TCHAR* pszLastRainfallSection{_T("LastRainfall")};
  const UINT nLastRainfallCount{GetProfileInt(pszLastRainfallSection, _T("Count"), INT_MAX)};
  if (nLastRainfallCount != INT_MAX)
  {
    m_nLastRainfallCount = nLastRainfallCount;
    Log(_T("Reloading last rainfall count of %d\n"), m_nLastRainfallCount);

#pragma warning(suppress: 26472)
    const auto nLastRainfallYear{static_cast<int32_t>(GetProfileInt(pszLastRainfallSection, _T("Year"), 0))};
#pragma warning(suppress: 26472)
    const auto nLastRainfallMonth{static_cast<uint16_t>(GetProfileInt(pszLastRainfallSection, _T("Month"), 0))};
#pragma warning(suppress: 26472)
    const auto nLastRainfallDay{static_cast<uint16_t>(GetProfileInt(pszLastRainfallSection, _T("Day"), 0))};
#pragma warning(suppress: 26472)
    const auto nLastRainfallHour{static_cast<uint16_t>(GetProfileInt(pszLastRainfallSection, _T("Hour"), 0))};
#pragma warning(suppress: 26472)
    const auto nLastRainfallMinute{static_cast<uint16_t>(GetProfileInt(pszLastRainfallSection, _T("Minute"), 0))};
#pragma warning(suppress: 26472)
    const auto nLastRainfallSecond{static_cast<uint16_t>(GetProfileInt(pszLastRainfallSection, _T("Second"), 0))};

    if ((nLastRainfallYear != 0) && (nLastRainfallMonth != 0) && (nLastRainfallDay != 0))
    {
      m_LastRainfallCount = DTimePlus::CLDate<DTimePlus::CGregorianCalendar>{nLastRainfallYear, nLastRainfallMonth, nLastRainfallDay,
                                                                             nLastRainfallHour, nLastRainfallMinute, nLastRainfallSecond};
      Log(_T("Reloading last rainfall time of %s\n"), m_LastRainfallCount.Format(_T("%x")).c_str());
    }
  }

  //Open the COM port if required
  if (m_nCOMPort != 0)
  {
    //Call CreateFile to open the comms port
    ATL::CString sPort;
    sPort.Format(_T("\\\\.\\COM%d"), m_nCOMPort);
    Log(_T("Opening COM port: %d\n"), m_nCOMPort);
    HANDLE hComm{CreateFile(sPort, GENERIC_READ | GENERIC_WRITE, 0, nullptr, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, nullptr)};
    if (hComm == INVALID_HANDLE_VALUE)
    {
      const DWORD dwLastError{GetLastError()};
      Log(_T("Unable to open specified COM port:%d, Error:%u\n"), m_nCOMPort, dwLastError);
      TerminateService(dwLastError);
    }
    m_Port.Attach(hComm);

    //Get the current state prior to changing it
    DCB dcb{};
    dcb.DCBlength = sizeof(DCB);
    if (!GetCommState(m_Port, &dcb))
    {
      const DWORD dwLastError{GetLastError()};
      Log(_T("Unable to get COM state, Error:%u\n"), dwLastError);
      TerminateService(dwLastError);
    }

    //Setup the various DCB settings
    dcb.BaudRate = m_nBaudRate;
    dcb.fOutX = FALSE;
    dcb.fInX = FALSE;
    dcb.Parity = NOPARITY;
    dcb.ByteSize = DATABITS_8;
    dcb.StopBits = ONESTOPBIT;
    dcb.fDtrControl = DTR_CONTROL_DISABLE;
    dcb.fRtsControl = RTS_CONTROL_DISABLE;

    //set the state of the port
    if (!SetCommState(m_Port, &dcb))
    {
      const DWORD dwLastError{GetLastError()};
      Log(_T("Unable to set COM state, Error:%u\n"), dwLastError);
      TerminateService(dwLastError);
    }

    //Set the timeouts for the port
    COMMTIMEOUTS ct{};
    ct.ReadIntervalTimeout = MAXDWORD;
    if (!SetCommTimeouts(m_Port, &ct))
    {
      const DWORD dwLastError{GetLastError()};
      Log(_T("Unable to set timeouts on COM port, Error:%u\n"), dwLastError);
      TerminateService(dwLastError);
    }

    //Setup the event mask for the port
    if (!SetCommMask(m_Port, EV_RXCHAR))
    {
      const DWORD dwLastError{GetLastError()};
      Log(_T("Unable to set event mask on COM port, Error:%u\n"), dwLastError);
      TerminateService(dwLastError);
    }
  }
  else
  {
    //Open up the YKUSH XS device if required
    if (m_bResetUsingYKUSHXS)
    {
      HID::StringArray deviceNames;
      Log(_T("Enumerating YKUSH XS devices connected to this computer\n"));
      if (!HID::CYKUSHXSDevice::Enumerate(deviceNames))
      {
        Log(_T("Unable to enumerate YKUSH XS devices, Error:%u\n"), GetLastError());
        TerminateService(ERROR_MOD_NOT_FOUND);
      }
      HID::String sDevicePath;
      for (const auto& device : deviceNames)
      {
        if (sDevicePath.length() == 0)
  #pragma warning(suppress: 26486 26489)
          sDevicePath = device;
  #pragma warning(suppress: 26489)
        Log(_T("Found YKUSH XS device with device path of %s\n"), device.c_str());
      }
      if (deviceNames.size() == 0)
      {
        Log(_T("No YKUSH XS devices found\n"));
        TerminateService(ERROR_FILE_NOT_FOUND);
      }

      //Open up the YKUSH XS device
  #pragma warning(suppress: 26446)
      Log(_T("Opening YKUSH XS device '%s'\n"), deviceNames[0].c_str());
  #pragma warning(suppress: 26446)
      if (!m_YKUSHFXDevice.Open(deviceNames[0].c_str()))
      {
        const DWORD dwError{GetLastError()};
  #pragma warning(suppress: 26446)
        Log(_T("Unable to open YKUSH XS device '%s', Error:%u\n"), deviceNames[0].c_str(), dwError);
        TerminateService(dwError);
      }

      Log(_T("Powering up USB port using YKUSH XS\n"));
      m_YKUSHFXDevice.PortUp();
    }

    Log(_T("Enumerating Fine Offset weather station devices connected to this computer\n"));
    HID::StringArray deviceNames;
    if (!CFOStation::Enumerate(deviceNames))
    {
      Log(_T("Unable to enumerate Fine Offset weather station devices, Error:%u\n"), GetLastError());
      TerminateService(ERROR_MOD_NOT_FOUND);
    }
    for (const auto& device : deviceNames)
    {
      if (m_sDevicePath.length() == 0)
  #pragma warning(suppress: 26486 26489)
        m_sDevicePath = device;
  #pragma warning(suppress: 26489)
      Log(_T("Found Fine Offset weather station with device path of %s\n"), device.c_str());
    }
    if (deviceNames.size() == 0)
    {
      Log(_T("No Fine Offset weather station devices found\n"));
      TerminateService(ERROR_FILE_NOT_FOUND);
    }

    //Open up the HID Fine Offset device
    if (!OpenWeatherStationAsHIDDevice())
    {
      const DWORD dwError{GetLastError()};
  #pragma warning(suppress: 26446)
      Log(_T("Unable to open weather station device '%s', Error:%u\n"), deviceNames[0].c_str(), dwError);
      TerminateService(dwError);
    }
  }

  //Connect to the DB
  if (!OpenDB(true))
    TerminateService(ERROR_BAD_NETPATH);

  ReportStatus(SERVICE_RUNNING);

  if (m_nCOMPort == 0)
  {
    //Synchronize with the station
    if (!SynchronizeWithStation())
      TerminateService(ERROR_READ_FAULT);

    //Next process all the historical data
    if (!ReadHistoricalData())
      TerminateService(ERROR_READ_FAULT);

    int nPreviousWriteAddress{-1};
    bool bWantStop{false};
    while (!bWantStop)
    {
      //Resynchronize with the station every 24 hours
      now = DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentLocalTime();
      if (((now - m_StationClock) > DTimePlus::CLTimeSpan{1, 0, 0, 0}))
        SynchronizeWithStation();
      else
      {
        //Check that we are not within N seconds of the station updating
        if (!SafeToReadFromDevice())
        {
          if (m_bDebugAuditing)
            Log(_T("Skipping read because not safe to do so at this time\n"));
        }
        else
        {
          //Get the address of the current write location from the device
          std::vector<BYTE> data;
          if (m_Device.ReadAddress(0x0, data, m_dwDeviceTimeout))
          {
  #pragma warning(suppress: 26446)
            const int nCurrentWriteAddress{(data[CFOStation::CURRENT_LOCATION_ADDRESSMSB] * 256) + data[CFOStation::CURRENT_LOCATION_ADDRESSLSB]};

            //Determine if the current write address is valid
            bool bValidCurrentWriteAddress{false};
            switch (m_StationType)
            {
              case FOStationType::WithSolarData:
              {
                //Check that the current write address is reasonable
                bValidCurrentWriteAddress = ((nCurrentWriteAddress >= CFOStation::MIN_LOCATION_ADDRESS) && (nCurrentWriteAddress <= CFOStation::MAX_SOLAR_LOCATION_ADDRESS));
                if (bValidCurrentWriteAddress && (nPreviousWriteAddress != -1))
                {
                  if ((nCurrentWriteAddress != nPreviousWriteAddress) && (nPreviousWriteAddress != CFOStation::MAX_SOLAR_LOCATION_ADDRESS))
                    bValidCurrentWriteAddress = (((nCurrentWriteAddress - nPreviousWriteAddress) % CFORecord::WITH_SOLAR_RECORDSIZE) == 0);
                }
                break;
              }
              case FOStationType::WithourSolarData:
              {
                //Check that the current write address is reasonable
                bValidCurrentWriteAddress = ((nCurrentWriteAddress >= CFOStation::MIN_LOCATION_ADDRESS) && (nCurrentWriteAddress <= CFOStation::MAX_WITHOUTSOLAR_LOCATION_ADDRESS));
                if (bValidCurrentWriteAddress && (nPreviousWriteAddress != -1))
                {
                  if ((nCurrentWriteAddress != nPreviousWriteAddress) && (nPreviousWriteAddress != CFOStation::MAX_WITHOUTSOLAR_LOCATION_ADDRESS))
                    bValidCurrentWriteAddress = (((nCurrentWriteAddress - nPreviousWriteAddress) % CFORecord::WITHOUT_SOLAR_RECORDSIZE) == 0);
                }
                break;
              }
              default:
              {
  #pragma warning(suppress: 26477)
                ATLASSERT(FALSE);
                break;
              }
            }

            //Check that the current write address is valid
            if (!bValidCurrentWriteAddress && m_bDebugAuditing)
              Log(_T("Current write address is 0x%04X which is not valid. Will skip processing this data.\n"), nCurrentWriteAddress);
            else if ((nPreviousWriteAddress != -1) && (nCurrentWriteAddress != nPreviousWriteAddress)) //See if the current write location has changed
            {
              if (m_bDebugAuditing)
                Log(_T("Processing new data at address 0x%04X\n"), nCurrentWriteAddress);
              if (m_Device.ReadAddress(nCurrentWriteAddress, data, m_dwDeviceTimeout))
              {
                switch (m_StationType)
                {
                  case FOStationType::WithSolarData:
                  {
                    if (data.size() > CFORecord::WITH_SOLAR_RECORDSIZE)
                      data.resize(CFORecord::WITH_SOLAR_RECORDSIZE);
                    break;
                  }
                  case FOStationType::WithourSolarData:
                  {
                    if (data.size() > CFORecord::WITHOUT_SOLAR_RECORDSIZE)
                      data.resize(CFORecord::WITHOUT_SOLAR_RECORDSIZE);
                    break;
                  }
                  default:
                  {
  #pragma warning(suppress: 26477)
                    ATLASSERT(FALSE);
                    break;
                  }
                }

                //parse and process the new data record
                CFORecord record;
                if (record.ParseFromUSBHID(data))
                {
                  record.m_LocalDateTime = DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentLocalTime();
                  record.m_UTCDateTime = DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentSystemTime();
                  if (m_bDebugAuditing)
                    Log(_T("Processing record at address 0x%04X for %s\n"), nCurrentWriteAddress, record.m_LocalDateTime.Format(_T("%x")).c_str());
                  ProcessCurrentRecord(record);
                }
                else
                  Log(_T("Could not parse new data at address 0x%04X\n"), nCurrentWriteAddress);
              }
              else
              {
                //PJ TO DO: To do email notification on failure to read here

                const DWORD dwLastError{GetLastError()};
                LogWin32Error(dwLastError, _T("An error occurred reading new data at address 0x%04X, Error:<WIN32ERROR>\n"), nCurrentWriteAddress);
                if (m_bResetUsingYKUSHXS && (dwLastError == ERROR_TIMEOUT))
                  ResetUSBWithYKUSHXS();

                //Close down the weather station handle
                Log(_T("Closing weather station device\n"));
                m_Device.Close();
              }
            }

            //Remember the write location if it was valid for the next time around
            if (bValidCurrentWriteAddress)
              nPreviousWriteAddress = nCurrentWriteAddress;
          }
          else
          {
            //PJ TO DO: To do email notification on failure to read here
            const DWORD dwLastError{GetLastError()};
            LogWin32Error(dwLastError, _T("An error occurred reading address 0x0, Error:<WIN32ERROR>\n"));
            if (m_bResetUsingYKUSHXS && (dwLastError == ERROR_TIMEOUT))
              ResetUSBWithYKUSHXS();

            //Close down the weather station handle
            Log(_T("Closing weather station device\n"));
            m_Device.Close();
          }
        }
      }

      //Reopen the device (and synchronize to it) if required
      if (m_Device.operator HANDLE() == INVALID_HANDLE_VALUE)
      {
        if (OpenWeatherStationAsHIDDevice())
          SynchronizeWithStation();
      }

      //Prepare for the next time around
      bWantStop = WaitForSingleObject(m_StopEvent, 10000) == WAIT_OBJECT_0;
    }
  }
  else
  {
    //Setup the OVERLAPPED structure we will need
    OVERLAPPED overlapped;
    memset(&overlapped, 0, sizeof(overlapped));
    overlapped.hEvent = m_COMPortDataEvent;

    //Loop around waiting for something to do
    CFORecord lastWeatherRecord;
    CFORecord lastLightRecord;
    ATL::CStringA sBuffer;
    std::array<HANDLE, 2> handles{m_StopEvent, m_COMPortDataEvent};
    bool bWantStop{false};
    while (!bWantStop)
    {
      DWORD dwEventMask{0};
  #pragma warning(suppress: 26486)
      if (WaitCommEvent(m_Port, &dwEventMask, &overlapped))
        ReadCOMPortData(sBuffer, overlapped, lastWeatherRecord, lastLightRecord);
      const DWORD dwLastError{GetLastError()};
      if (dwLastError == ERROR_IO_PENDING)
      {
        //If the read method failed synchronously then wait on the 2 events for the read to complete
  #pragma warning(suppress: 26485)
        const DWORD dwWait{WaitForMultipleObjects(2, handles.data(), FALSE, 300000)};
        if (dwWait == WAIT_TIMEOUT)
        {
          CancelIo(m_Port);
          //PJ TO DO: To do email notification on waiting too long for some serial data
          Log(_T("Have not read data from the COM port in 5 minutes\n"));
        }
        else if (dwWait != WAIT_FAILED)
        {
          //Work out which event got signalled
          const DWORD dwSignaledHandle{dwWait - WAIT_OBJECT_0};
          if (dwSignaledHandle == 0)
            bWantStop = true;
          else
            ReadCOMPortData(sBuffer, overlapped, lastWeatherRecord, lastLightRecord);
        }
      }
      else
      {
        //PJ TO DO: To do email notification on failure here
        Log(_T("Unable to read data from the COM port, Error:%u\n"), dwLastError);
        TerminateService(dwLastError);
      }
    }
  }

  Log(_T("******* Service stopping *******\n"));
}

void CMyService::ShowHelp() noexcept
{
  _tprintf(_T("Fine Offset Weather Station\nUsage: FOStation [ -install | -uninstall | -remove | -start | -stop | -help | -? ]\n"));
}

bool CMyService::UploadToWOW(_In_ const CFORecord& record)
{
  bool bUpload{false};
  const DTimePlus::CLDate<DTimePlus::CGregorianCalendar> now{DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentSystemTime()};
  if (m_LastMETWOWUpload.IsValid())
  {
    if (m_nMETWOWLoggingInterval == 0)
      bUpload = true;
    else
    {
      const DTimePlus::CLTimeSpan timeSpan{now - m_LastMETWOWUpload};
      bUpload = (timeSpan.SecondsAsDouble() > (m_nMETWOWLoggingInterval * 60.0));
    }
  }
  else
    bUpload = true;
  if (!bUpload)
    return false;

  if (m_bDebugAuditing)
    Log(_T("Uploading record to MET WOW\n"));

  //Form the request string we will be sending
  SYSTEMTIME st;
  GetSystemTime(&st);
  ATL::CString sDateUTC;
  sDateUTC.Format(_T("%04d-%02d-%02d+%02d%%3A%02d%%3A%02d"), static_cast<int>(st.wYear), static_cast<int>(st.wMonth), static_cast<int>(st.wDay), static_cast<int>(st.wHour), static_cast<int>(st.wMinute), static_cast<int>(st.wSecond));
  ATL::CString sRequest;
  sRequest.Format(_T("/automaticreading?siteid=%s&siteAuthenticationKey=%s&dateutc=%s&softwaretype=FOStation%s"), m_sMETWOWSiteID.GetString(), m_sMETWOWSiteAuthenticationKey.GetString(), sDateUTC.GetString(), m_sVersionInfo.GetString());
  bool bValidRequest{false};
  if (record.m_nOutdoorRelativeHumidity.has_value())
  {
    sRequest.AppendFormat(_T("&humidity=%d"), record.m_nOutdoorRelativeHumidity.value());
    bValidRequest = true;
  }
  if (record.m_fOutdoorTemperature.has_value())
  {
    sRequest.AppendFormat(_T("&tempf=%.1f"), (static_cast<double>(record.m_fOutdoorTemperature.value()) * 9 / 5) + 32);
    bValidRequest = true;
  }
  if (record.m_fAbsolutePressure.has_value())
  {
    sRequest.AppendFormat(_T("&baromin=%.3f"), record.m_fAbsolutePressure.value() * 0.02953);
    bValidRequest = true;
  }
  if (record.m_fAverageWindSpeed.has_value())
  {
    sRequest.AppendFormat(_T("&windspeedmph=%.1f"), record.m_fAverageWindSpeed.value() * 2.23694);
    bValidRequest = true;
  }
  if (record.m_fGustWindSpeed.has_value())
  {
    sRequest.AppendFormat(_T("&windgustmph=%.1f"), record.m_fGustWindSpeed.value() * 2.23694);
    bValidRequest = true;
  }
  if (record.m_nWindDirection.has_value())
  {
    sRequest.AppendFormat(_T("&winddir=%d"), record.m_nWindDirection.value());
    bValidRequest = true;
  }
  if (record.m_fRainfallSinceMidnight.has_value())
  {
    sRequest.AppendFormat(_T("&dailyrainin=%.3f"), record.m_fRainfallSinceMidnight.value() * 0.0393701);
    bValidRequest = true;
  }
  if (record.m_fOutdoorTemperature.has_value() && record.m_nOutdoorRelativeHumidity.has_value())
  {
    const double fDewPoint{DewPoint(record.m_fOutdoorTemperature.value(), record.m_nOutdoorRelativeHumidity.value())};
    sRequest.AppendFormat(_T("&dewptf=%.1f"), (fDewPoint * 9 / 5) + 32);
    bValidRequest = true;
  }
  if (!bValidRequest)
  {
    Log(_T("No actual data available to upload to MET WOW\n"));
    return false;
  }
  if (m_bDebugAuditing)
  {
    ATL::CString sRequestForLogging{sRequest};
    ATL::CString sAuthenticationKeyToReplace{_T("siteAuthenticationKey=")};
    sAuthenticationKeyToReplace += m_sMETWOWSiteAuthenticationKey;
    sRequestForLogging.Replace(sAuthenticationKeyToReplace, _T("siteAuthenticationKey=******"));
    Log(_T("MET WOW request is %s\n"), sRequestForLogging.GetString());
  }

  //Create the internet session
  WinHTTPWrappers::CSession session;
#pragma warning(suppress: 26477)
  HRESULT hr{session.Initialize(L"FOStation", WINHTTP_ACCESS_TYPE_DEFAULT_PROXY, WINHTTP_NO_PROXY_NAME, WINHTTP_NO_PROXY_BYPASS, 0)};
  if (FAILED(hr))
  {
    Log(_T("Failed to create MET WOW session, Error:0x%08X\n"), hr);
    return false;
  }

  //Create the internet connection
  WinHTTPWrappers::CConnection connection;
  hr = connection.Initialize(session, L"wow.metoffice.gov.uk", 80);
  if (FAILED(hr))
  {
    Log(_T("Failed to connect to MET WOW web site, Error:0x%08X\n"), hr);
    return false;
  }

  //Create and issue the request
  WinHTTPWrappers::CSyncDownloader file;
  hr = file.Initialize(connection, ATL::CT2W{sRequest});
  if (FAILED(hr))
  {
    Log(_T("Failed to create MET WOW file, Error:0x%08X\n"), hr);
    return false;
  }

  //Do the download
  hr = file.SendRequestSync();
  file.ReleaseResources();
  if (FAILED(hr))
  {
#pragma warning(suppress: 26472 26490)
    ATL::CStringA sResponse{reinterpret_cast<const char*>(file.m_Response.data()), static_cast<int>(file.m_Response.size())};
    bool bStatusCodeValid{false};
    const DWORD dwLastStatusCode{file.GetLastStatusCode(bStatusCodeValid)};
    if (bStatusCodeValid)
      Log(_T("Failed to send MET WOW request, Error:0x%08X, response:%s, Status code:%u\n"), hr, ATL::CA2T{sResponse}.operator LPTSTR(), dwLastStatusCode);
    else
      Log(_T("Failed to send MET WOW request, Error:0x%08X, response:%s\n"), hr, ATL::CA2T{sResponse}.operator LPTSTR());
    return false;
  }

  //Remember the time of this last upload
  m_LastMETWOWUpload = DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentSystemTime();

  return true;
}

bool CMyService::UploadToWeatherUnderground(_In_ const CFORecord& record)
{
  bool bUpload{false};
  const DTimePlus::CLDate<DTimePlus::CGregorianCalendar> now{DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentSystemTime()};
  if (m_LastWeatherUndergroundUpload.IsValid())
  {
    if (m_nWeatherUndergroundLoggingInterval == 0)
      bUpload = true;
    else
    {
      const DTimePlus::CLTimeSpan timeSpan{now - m_LastWeatherUndergroundUpload };
      bUpload = (timeSpan.SecondsAsDouble() > (m_nWeatherUndergroundLoggingInterval * 60.0));
    }
  }
  else
    bUpload = true;
  if (!bUpload)
    return false;

  if (m_bDebugAuditing)
    Log(_T("Uploading record to Weather Underground\n"));

  //Form the request string we will be sending
  SYSTEMTIME st;
  GetSystemTime(&st);
  ATL::CString sDateUTC;
  sDateUTC.Format(_T("%04d-%02d-%02d+%02d%%3A%02d%%3A%02d"), static_cast<int>(st.wYear), static_cast<int>(st.wMonth), static_cast<int>(st.wDay), static_cast<int>(st.wHour), static_cast<int>(st.wMinute), static_cast<int>(st.wSecond));
  ATL::CString sRequest;
  sRequest.Format(_T("/weatherstation/updateweatherstation.php?ID=%s&PASSWORD=%s&dateutc=%s&action=updateraw&softwaretype=FOStation%s"), m_sWeatherUndergroundID.GetString(), m_sWeatherUndergroundPassword.GetString(), sDateUTC.GetString(), m_sVersionInfo.GetString());
  bool bValidRequest{false};
  if (record.m_nOutdoorRelativeHumidity.has_value())
  {
    sRequest.AppendFormat(_T("&humidity=%d"), record.m_nOutdoorRelativeHumidity.value());
    bValidRequest = true;
  }
  if (record.m_fOutdoorTemperature.has_value())
  {
    sRequest.AppendFormat(_T("&tempf=%.1f"), (static_cast<double>(record.m_fOutdoorTemperature.value()) * 9 / 5) + 32);
    bValidRequest = true;
  }
  if (record.m_fIndoorTemperature.has_value())
  {
    sRequest.AppendFormat(_T("&indoortempf=%.1f"), (static_cast<double>(record.m_fIndoorTemperature.value()) * 9 / 5) + 32);
    bValidRequest = true;
  }
  if (record.m_nIndoorRelativeHumidity.has_value())
  {
    sRequest.AppendFormat(_T("&indoorhumidity=%d"), record.m_nIndoorRelativeHumidity.value());
    bValidRequest = true;
  }
  if (record.m_fAbsolutePressure.has_value())
  {
    sRequest.AppendFormat(_T("&baromin=%.3f"), record.m_fAbsolutePressure.value() * 0.02953);
    bValidRequest = true;
  }
  if (record.m_fAverageWindSpeed.has_value())
  {
    sRequest.AppendFormat(_T("&windspeedmph=%.1f"), record.m_fAverageWindSpeed.value() * 2.23694);
    bValidRequest = true;
  }
  if (record.m_fGustWindSpeed.has_value())
  {
    sRequest.AppendFormat(_T("&windgustmph=%.1f"), record.m_fGustWindSpeed.value() * 2.23694);
    bValidRequest = true;
  }
  if (record.m_nWindDirection.has_value())
  {
    sRequest.AppendFormat(_T("&winddir=%d"), record.m_nWindDirection.value());
    bValidRequest = true;
  }
  if (record.m_fRainfallSinceMidnight.has_value())
  {
    sRequest.AppendFormat(_T("&dailyrainin=%.3f"), record.m_fRainfallSinceMidnight.value() * 0.0393701);
    bValidRequest = true;
  }
  if (record.m_fOutdoorTemperature.has_value() && record.m_nOutdoorRelativeHumidity.has_value())
  {
    const double fDewPoint{DewPoint(record.m_fOutdoorTemperature.value(), record.m_nOutdoorRelativeHumidity.value())};
    sRequest.AppendFormat(_T("&dewptf=%.1f"), (fDewPoint * 9 / 5) + 32);
    bValidRequest = true;
  }
  if (record.m_nUV.has_value())
  {
    sRequest.AppendFormat(_T("&UV=%d"), record.m_nUV.value());
    bValidRequest = true;
  }
  if (record.m_fIlluminance.has_value())
  {
    sRequest.AppendFormat(_T("&solarradiation=%.3f"), record.m_fIlluminance.value() * 0.0079);
    bValidRequest = true;
  }

  if (!bValidRequest)
  {
    Log(_T("No actual data available to upload to Weather Underground\n"));
    return false;
  }
  if (m_bDebugAuditing)
  {
    ATL::CString sRequestForLogging{sRequest};
    ATL::CString sPasswordToReplace{_T("PASSWORD=")};
    sPasswordToReplace += m_sWeatherUndergroundPassword;
    sRequestForLogging.Replace(sPasswordToReplace, _T("PASSWORD=******"));
    Log(_T("Weather Underground request is %s\n"), sRequestForLogging.GetString());
  }

  //Create the internet session
  WinHTTPWrappers::CSession session;
#pragma warning(suppress: 26477)
  HRESULT hr{session.Initialize(L"FOStation", WINHTTP_ACCESS_TYPE_DEFAULT_PROXY, WINHTTP_NO_PROXY_NAME, WINHTTP_NO_PROXY_BYPASS, 0)};
  if (FAILED(hr))
  {
    Log(_T("Failed to create Weather Underground session, Error:0x%08X\n"), hr);
    return false;
  }

  //Create the internet connection
  WinHTTPWrappers::CConnection connection;
  hr = connection.Initialize(session, L"weatherstation.wunderground.com", 443);
  if (FAILED(hr))
  {
    Log(_T("Failed to connect to Weather Underground web site, Error:0x%08X\n"), hr);
    return false;
  }

  //Create and issue the request
  WinHTTPWrappers::CSyncDownloader file;
  hr = file.Initialize(connection, ATL::CT2W{sRequest}, nullptr, nullptr, WINHTTP_NO_REFERER, WINHTTP_DEFAULT_ACCEPT_TYPES, WINHTTP_FLAG_SECURE);
  if (FAILED(hr))
  {
    Log(_T("Failed to create Weather Underground file, Error:0x%08X\n"), hr);
    return false;
  }

  //Do the download
  hr = file.SendRequestSync();
  file.ReleaseResources();
  if (FAILED(hr))
  {
#pragma warning(suppress: 26472 26490)
    ATL::CStringA sResponse{reinterpret_cast<const char*>(file.m_Response.data()), static_cast<int>(file.m_Response.size())};
    bool bStatusCodeValid{false};
    const DWORD dwLastStatusCode{file.GetLastStatusCode(bStatusCodeValid)};
    if (bStatusCodeValid)
      Log(_T("Failed to send Weather Underground request, Error:0x%08X, response:%s, Status code:%u\n"), hr, ATL::CA2T{sResponse}.operator LPTSTR(), dwLastStatusCode);
    else
      Log(_T("Failed to send Weather Underground request, Error:0x%08X, response:%s\n"), hr, ATL::CA2T{sResponse}.operator LPTSTR());
    return false;
  }

  //Remember the time of this last upload
  m_LastWeatherUndergroundUpload = DTimePlus::CLDate<DTimePlus::CGregorianCalendar>::CurrentSystemTime();

  return true;
}

//Use the Davis algorithm for Dewpoint from https://www.davisinstruments.com/product_documents/weather/app_notes/AN_28-derived-weather-variables.pdf
double CMyService::DewPoint(_In_ double fTemperature, _In_ double fRelativeHumidity) noexcept
{
  const double fLnVapour{log(fRelativeHumidity * 0.01 * 6.112 * exp((17.62 * fTemperature) / (fTemperature + 243.12)))};
  const double fDewPoint{((243.12 * fLnVapour) - 440.1) / (19.43 - fLnVapour)};
  if (fDewPoint > fTemperature)
    return fTemperature;
  else
    return fDewPoint;
}

int _tmain()
{
  //All that is required to get the service up and running
  CNTServiceCommandLineInfo cmdInfo;
  theService.SetAllowCommandLine(TRUE);
#pragma warning(suppress: 26486)
  theService.ParseCommandLine(cmdInfo);
  return theService.ProcessShellCommand(cmdInfo);
}
