/*
Module : FOStationWrappers.h
Purpose: Defines the interface for a set of C++ classes which encapsulates the FineOffset weather stations using the Windows 
         HID APIs or via an Arduino Uno/Micro/Mega sketch which uses a AUREL RX-4MM5 AM receiver to decode the FineOffset
         packets over the air.
History: PJN / 07-09-2019 1. Initial release
         PJN / 24-07-2022 1. Updated the code to handle the COM port formats from the Arduino FineOffsetRFWeather.ino sketch
                          which is available from http://www.naughter.com/download/FineOffsetRFWeather.ino.
                          2. Updated the code to use C++ uniform initialization for all variable declarations.

Copyright (c) 2019 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#ifndef __FOSTATIONWRAPPERS_H__
#define __FOSTATIONWRAPPERS_H__

#ifndef CFOSTATIONWRAPPERS_EXT_CLASS
#define CFOSTATIONWRAPPERS_EXT_CLASS
#endif //#ifndef CFOSTATIONWRAPPERS_EXT_CLASS


#pragma comment(lib, "hid.lib")
#pragma comment(lib, "Setupapi.lib")


//////////////////// Includes /////////////////////////////////////////////////

#include "HIDWrappers.h"
#include "DTime+JulianCalendar.h"
#include "DTime+GregorianCalendar.h"
#include "DTime+HebrewCalendar.h"
#include "DTime+CDate.h"
#include "DTime+CLDate.h"
#include <optional>

#ifndef _ARRAY_
#pragma message("To avoid this message, please put array in your pre compiled header (normally stdafx.h)")
#include <array>
#endif //#ifndef _ARRAY_


//////////////////// Classes //////////////////////////////////////////////////

//Class which represents a record for a FO weather station (supports both record with and without solar data)
class CFOSTATIONWRAPPERS_EXT_CLASS CFORecord
{
public:
//Enums
  enum
  {
    WITHOUT_SOLAR_RECORDSIZE = 0X10,
    WITH_SOLAR_RECORDSIZE = 0X14
  };

  //Constructors / Destructors
  CFORecord() noexcept : m_nSampleTime(0),
                         m_bRainfallOverflowFlag(false)
  {
  }
  CFORecord(const CFORecord&) = default;
  CFORecord(CFORecord&&) = default;
  ~CFORecord() = default;

  //Methods
  CFORecord& operator=(const CFORecord&) = default;
  CFORecord& operator=(CFORecord&&) = default;
  bool ParseFromUSBHID(std::vector<BYTE>& data)
  {
    const auto nRecordSize{data.size()};
    if ((nRecordSize != WITHOUT_SOLAR_RECORDSIZE) && (nRecordSize != WITH_SOLAR_RECORDSIZE))
      return false;
#pragma warning(suppress: 26446)
    m_nSampleTime = data[0];
#pragma warning(suppress: 26446)
    if (data[1] != 0xFF)
#pragma warning(suppress: 26446)
      m_nIndoorRelativeHumidity = data[1];
    else
      m_nIndoorRelativeHumidity.reset();
#pragma warning(suppress: 26446)
    const int nIndoorTemperature{data[2] + (data[3] * 256)};
    if (nIndoorTemperature != 0xFFFF)
    {
#pragma warning(suppress: 26446)
      m_fIndoorTemperature = static_cast<float>((static_cast<double>(data[2]) + static_cast<double>(data[3] & 0x7F) * 256) / 10.0);
#pragma warning(suppress: 26446)
      if (data[3] & 0x80)
        m_fIndoorTemperature = -m_fIndoorTemperature.value();
    }
    else
      m_fIndoorTemperature.reset();
#pragma warning(suppress: 26446)
    if (data[4] != 0xFF)
#pragma warning(suppress: 26446)
      m_nOutdoorRelativeHumidity = data[4];
    else
      m_nOutdoorRelativeHumidity.reset();
#pragma warning(suppress: 26446)
    const int nOutdoorTemperature{data[5] + (data[6] * 256)};
    if (nOutdoorTemperature != 0xFFFF)
    {
#pragma warning(suppress: 26446)
      m_fOutdoorTemperature = static_cast<float>((static_cast<double>(data[5]) + static_cast<double>(data[6] & 0x7F) * 256) / 10.0);
#pragma warning(suppress: 26446)
      if (data[6] & 0x80)
        m_fOutdoorTemperature = -m_fOutdoorTemperature.value();
    }
    else
      m_fOutdoorTemperature.reset();
#pragma warning(suppress: 26446)
    const int nAbsolutePressure{data[7] + (data[8] * 256)};
    if (nAbsolutePressure != 0xFFFF)
      m_fAbsolutePressure = static_cast<float>(nAbsolutePressure / 10.0);
    else
      m_fAbsolutePressure.reset();
#pragma warning(suppress: 26446)
    const int nAverageWindspeed{data[9] + (data[11] & 0x0F) * 256};
    if (nAverageWindspeed != 0xFFFF)
#pragma warning(suppress: 26446)
      m_fAverageWindSpeed = static_cast<float>((static_cast<double>(data[9]) + static_cast<double>(data[11] & 0x0F) * 256) / 10.0);
    else
      m_fAverageWindSpeed.reset();
#pragma warning(suppress: 26446)
    const int nGustWindspeed{data[10] + ((data[11] & 0xF0) >> 4) * 256};
    if (nGustWindspeed != 0xFFFF)
#pragma warning(suppress: 26446)
      m_fGustWindSpeed = static_cast<float>((static_cast<double>(data[10]) + static_cast<double>(data[11] & 0xF0) * 16) / 10.0);
    else
      m_fGustWindSpeed.reset();
#pragma warning(suppress: 26446)
    if (data[12] & 0x80)
      m_nWindDirection.reset();
    else
#pragma warning(suppress: 26446)
      m_nWindDirection = static_cast<int>(((data[12] & 0x0F) * 22.5) + 0.5);
#pragma warning(suppress: 26446)
    if (data[15] & 0x40)
      m_nTotalRain.reset();
    else
#pragma warning(suppress: 26446)
      m_nTotalRain = data[13] + (data[14] * 256);
#pragma warning(suppress: 26446)
    m_bRainfallOverflowFlag = data[15] & 0x80;

    if (nRecordSize == WITH_SOLAR_RECORDSIZE) //Illuminance and UV data is only provided with the solar sized record
    {
#pragma warning(suppress: 26446)
      const int nIlluminance{data[16] + (data[17] * 256) + (data[18] * 65536)};
      if (nIlluminance != 0xFFFFFF)
        m_fIlluminance = static_cast<float>(nIlluminance / 10.0);
      else
        m_fIlluminance.reset();
#pragma warning(suppress: 26446)
      if (data[19] != 0xFF)
#pragma warning(suppress: 26446)
        m_nUV = data[19];
      else
        m_nUV.reset();
    }
    else
    {
      m_fIlluminance.reset();
      m_nUV.reset();
    }

    return true;
  }

  int ParseFromCOMWeatherMsg(_In_ const ATL::CStringA& sLine)
  {
    int nValuesRead{0};
    int nCurPosition{0};
    ATL::CStringA sToken{sLine.Tokenize(",", nCurPosition) };
    while (sToken != "")
    {
      sToken.Trim();
      const int nColon{sToken.Find(":")};
      if (nColon != -1)
      {
        ATL::CStringA sKey{sToken.Mid(0, nColon)};
        ATL::CStringA sValue{sToken.Mid(nColon + 1)};
        if (sKey.CompareNoCase("DirW") == 0)
        {
          m_nWindDirection = atoi(sValue);
          ++nValuesRead;
        }
        else if (sKey.CompareNoCase("TempO") == 0)
        {
          m_fOutdoorTemperature = static_cast<float>(atof(sValue));
          ++nValuesRead;
        }
        else if (sKey.CompareNoCase("HumidityO") == 0)
        {
          m_nOutdoorRelativeHumidity = atoi(sValue);
          ++nValuesRead;
        }
        else if (sKey.CompareNoCase("TempI") == 0)
        {
          m_fIndoorTemperature = static_cast<float>(atof(sValue));
          ++nValuesRead;
        }
        else if (sKey.CompareNoCase("HumidityI") == 0)
        {
          m_nIndoorRelativeHumidity = static_cast<int>(atof(sValue) + 0.5);
          ++nValuesRead;
        }
        else if (sKey.CompareNoCase("AvgW") == 0)
        {
          m_fAverageWindSpeed = static_cast<float>(atof(sValue));
          ++nValuesRead;
        }
        else if (sKey.CompareNoCase("GustW") == 0)
        {
          m_fGustWindSpeed = static_cast<float>(atof(sValue));
          ++nValuesRead;
        }
        else if (sKey.CompareNoCase("Rain") == 0)
        {
          m_nTotalRain = atoi(sValue);
          ++nValuesRead;
        }
        else if (sKey.CompareNoCase("RSSI") == 0)
        {
          m_nRSSI = atoi(sValue);
          ++nValuesRead;
        }
        else if (sKey.CompareNoCase("Pressure") == 0)
        {
          m_fAbsolutePressure = static_cast<float>(atof(sValue));
          ++nValuesRead;
        }
        else if (sKey.CompareNoCase("BatteryStatus") == 0)
        {
          if (sValue.CompareNoCase("OK") == 0)
            m_bBatteryStatusOK = true;
          else
            m_bBatteryStatusOK = false;
          ++nValuesRead;
        }
      }

      //Prepare for the next time around
      sToken = sLine.Tokenize(",", nCurPosition);
    }
    return nValuesRead;
  }

  int ParseFromCOMLightMsg(_In_ const ATL::CStringA& sLine)
  {
    int nValuesRead{0};
    int nCurPosition{0};
    ATL::CStringA sToken{sLine.Tokenize(",", nCurPosition)};
    while (sToken != "")
    {
      sToken.Trim();
      const int nColon{sToken.Find(":")};
      if (nColon != -1)
      {
        ATL::CStringA sKey{sToken.Mid(0, nColon)};
        ATL::CStringA sValue{sToken.Mid(nColon + 1)};
        if (sKey.CompareNoCase("SensorID") == 0)
        {
          m_nSensorID = atoi(sValue);
          ++nValuesRead;
        }
        else if (sKey.CompareNoCase("SensorStatus") == 0)
        {
          if (sValue.CompareNoCase("OK") == 0)
            m_bLightStatusOK = true;
          else
            m_bLightStatusOK = false;
          ++nValuesRead;
        }
        else if (sKey.CompareNoCase("UV") == 0)
        {
          m_nUV = atoi(sValue);
          ++nValuesRead;
        }
        else if (sKey.CompareNoCase("RSSI") == 0)
        {
          m_nRSSI = atoi(sValue);
          ++nValuesRead;
        }
        else if (sKey.CompareNoCase("TempI") == 0)
        {
          m_fIndoorTemperature = static_cast<float>(atof(sValue));
          ++nValuesRead;
        }
        else if (sKey.CompareNoCase("Pressure") == 0)
        {
          m_fAbsolutePressure = static_cast<float>(atof(sValue));
          ++nValuesRead;
        }
        else if (sKey.CompareNoCase("HumidityI") == 0)
        {
          m_nIndoorRelativeHumidity = static_cast<int>(atof(sValue) + 0.5);
          ++nValuesRead;
        }
        else if (sKey.CompareNoCase("Lux") == 0)
        {
          m_fIlluminance = float{static_cast<float>(atof(sValue))};
          ++nValuesRead;
        }
      }

      //Prepare for the next time around
      sToken = sLine.Tokenize(",", nCurPosition);
    }
    return nValuesRead;
  }


//Member variables
  int m_nSampleTime; //The sampling time in minutes for this record
  std::optional<int> m_nIndoorRelativeHumidity; //The indoor relative humidity as a percentage
  std::optional<float> m_fIndoorTemperature; //The indoor temperature in Celsius
  std::optional<int> m_nOutdoorRelativeHumidity; //The outdoor relative humidity as a percentage
  std::optional<float> m_fOutdoorTemperature; //The outdoor temperature in Celsius
  std::optional<float> m_fAbsolutePressure; //The absolute pressure in hpa
  std::optional<float> m_fAverageWindSpeed; //The average wind speed in m/s
  std::optional<float> m_fGustWindSpeed; //The gust wind speed in m/s
  std::optional<int> m_nWindDirection; //The wind direction in degrees east of north (0-360)
  std::optional<int> m_nTotalRain; //The total rainfall counter value
  bool m_bRainfallOverflowFlag; //Has the total rainfall counter value overflowed
  std::optional<float> m_fIlluminance; //The illuminance value in lux
  std::optional<int> m_nUV; //The UV value
  std::optional<float> m_fRainfallSinceMidnight; //The rainfall in mm since midnight for this record
  std::optional<float> m_fRainfall; //The rainfall in mm for this record
  std::optional<bool> m_bBatteryStatusOK; //The battery status indicator
  std::optional<int> m_nRSSI; //The RSSI value
  std::optional<int> m_nSensorID; //The ID value from the Light message
  std::optional<bool> m_bLightStatusOK; //The light status indicator value
  DTimePlus::CLDate<DTimePlus::CGregorianCalendar> m_LocalDateTime; //The local date / time for this record
  DTimePlus::CLDate<DTimePlus::CGregorianCalendar> m_UTCDateTime; //The UTC date / time for this record
};


//Main class which encapsulates a FO weather station
class CFOSTATIONWRAPPERS_EXT_CLASS CFOStation : public HID::CDevice
{
public:
//Enums
  enum
  {
    CURRENT_LOCATION_ADDRESSLSB = 0X1E,
    CURRENT_LOCATION_ADDRESSMSB = 0X1F,
    SAMPLE_TIME_ADDRESS = 0X0,
    MIN_LOCATION_ADDRESS = 0xF0,
    MAX_SOLAR_LOCATION_ADDRESS = 0XFFEC,
    MAX_WITHOUTSOLAR_LOCATION_ADDRESS = 0XFFF0,
    HISTORICAL_RECORDS_ADDRESSLSB = 0X1B,
    HISTORICAL_RECORDS_ADDRESSMSB = 0X1C
  };

//Constructors / Destructors
  CFOStation() noexcept : m_Overlapped{}
  {
  }
  CFOStation(const CFOStation&) = delete;
  CFOStation(CFOStation&&) = delete;
  ~CFOStation() = default;

//Methods
  CFOStation& operator=(const CFOStation&) = delete;
  CFOStation& operator=(CFOStation&&) = delete;
  static bool Enumerate(_Out_ HID::StringArray& deviceNames, _In_ DWORD dwFlags = DIGCF_PRESENT | DIGCF_DEVICEINTERFACE, _In_opt_ LPCTSTR pszEnumerator = nullptr, _In_ USHORT idVendor = 0x1941, _In_ USHORT idProduct = 0x8021)
  {
    return EnumerateDevices(idVendor, idProduct, deviceNames, dwFlags, pszEnumerator);
  }

#pragma warning(suppress: 26434)
  bool Open(_In_ LPCTSTR pszDeviceName, _In_ DWORD dwDesiredAccess = GENERIC_READ | GENERIC_WRITE, _In_ DWORD dwShareMode = FILE_SHARE_READ | FILE_SHARE_WRITE, _In_opt_ LPSECURITY_ATTRIBUTES lpSecurityAttributes = nullptr, _In_ DWORD dwFlagsAndAttributes = 0) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_h == INVALID_HANDLE_VALUE);
#pragma warning(suppress: 26477)
    ATLASSERT(m_ReadEvent == nullptr);

    //Create the read event
    if (!m_ReadEvent.Create(nullptr, FALSE, FALSE, nullptr))
      return false;
    m_Overlapped.hEvent = m_ReadEvent;

    //Delegate to the base class
    return CDevice::Open(pszDeviceName, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwFlagsAndAttributes);
  }

  _Success_(return != 0) bool ReadInputReportTimeout(_Inout_ std::vector<BYTE>& inputReport, DWORD dwTimeout)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_h != INVALID_HANDLE_VALUE);

    DWORD dwNumberOfBytesRead{0};
#pragma warning(suppress: 26472)
    if (ReadFile(*this, inputReport.data(), static_cast<DWORD>(inputReport.size()), &dwNumberOfBytesRead, &m_Overlapped))
    {
      inputReport.resize(dwNumberOfBytesRead);
      return true;
    }
    if (GetLastError() != ERROR_IO_PENDING)
      return false;
    if (WaitForSingleObject(m_ReadEvent, dwTimeout) != WAIT_OBJECT_0)
    {
      CancelIo(*this);
      SetLastError(ERROR_TIMEOUT);
      return false;
    }
    const bool bSuccess{GetOverlappedResult(*this, &m_Overlapped, &dwNumberOfBytesRead, TRUE) != FALSE};
    if (!bSuccess)
      return false;
    inputReport.resize(dwNumberOfBytesRead);
    return true;
  }

  _Success_(return != 0) bool ReadInputReport(_Inout_ std::vector<BYTE>& inputReport)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_h != INVALID_HANDLE_VALUE);

    DWORD dwNumberOfBytesRead{0};
#pragma warning(suppress: 26472)
    const bool bSuccess{ReadFile(*this, inputReport.data(), static_cast<DWORD>(inputReport.size()), &dwNumberOfBytesRead, nullptr) != FALSE};
    if (!bSuccess)
      return false;
    inputReport.resize(dwNumberOfBytesRead);
    return true;
  }

  _Success_(return != 0) bool WriteFileTimeout(_In_reads_bytes_opt_(nNumberOfBytesToWrite) LPCVOID lpBuffer, _In_ DWORD nNumberOfBytesToWrite, _Out_ LPDWORD lpNumberOfBytesWritten, DWORD dwTimeout) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_h != INVALID_HANDLE_VALUE);

    //Issue the WriteFile call using the overlapped IO
    if (WriteFile(*this, lpBuffer, nNumberOfBytesToWrite, lpNumberOfBytesWritten, &m_Overlapped))
      return true;
    if (GetLastError() != ERROR_IO_PENDING)
      return false;
    if (WaitForSingleObject(m_ReadEvent, dwTimeout) != WAIT_OBJECT_0)
    {
      CancelIo(*this);
      return false;
    }
    return GetOverlappedResult(*this, &m_Overlapped, lpNumberOfBytesWritten, TRUE);
  }

  _Success_(return != 0) bool ReadAddress(_In_ int nAddress, _Inout_ std::vector<BYTE>& buffer, _In_ DWORD dwTimeout = 0, _In_ DWORD dwResponseDelay = 100)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_h != INVALID_HANDLE_VALUE);

    //Issue the read address request
#pragma warning(suppress: 26472)
    std::array<BYTE, 9> readRequest{0, 0xA1, static_cast<BYTE>(nAddress >> 8), static_cast<BYTE>(nAddress & 0xFF), 0x20, 0xA1, 0x00, 0x00, 0x20};

    bool bSuccess{false};
    DWORD dwWritten{0};
    if (dwTimeout)
#pragma warning(suppress: 26472)
      bSuccess = WriteFileTimeout(readRequest.data(), static_cast<DWORD>(readRequest.size()), &dwWritten, dwTimeout);
    else
#pragma warning(suppress: 26472)
      bSuccess = WriteFile(*this, readRequest.data(), static_cast<DWORD>(readRequest.size()), &dwWritten, nullptr);
    if (!bSuccess)
      return false;

    //Wait for the response
    Sleep(dwResponseDelay);

    //Retrieve the 32 bytes of the response in 4 HID input reports
    buffer.clear();
    for (int i{1}; i<=4; i++)
    {
      std::vector<BYTE> response{9, std::allocator<BYTE>{}};
      if (dwTimeout)
        bSuccess = ReadInputReportTimeout(response, dwTimeout);
      else
        bSuccess = ReadInputReport(response);
      if (!bSuccess)
        return false;
      buffer.insert(buffer.end(), response.begin() + 1, response.end()); //Drop the Report ID in the input report from the data we give back in the response
    }
    if (buffer.size() != 32)
    {
      SetLastError(ERROR_INVALID_DATA);
      return false;
    }
    return true;
  }

  _Success_(return != 0) bool WriteAddress(_In_ int nAddress, _In_ BYTE byValue, DWORD dwTimeout = 0)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_h != INVALID_HANDLE_VALUE);

    //Issue the write address request
#pragma warning(suppress: 26472)
    std::array<BYTE, 9> writeRequest{0, 0xA2, static_cast<BYTE>(nAddress >> 8), static_cast<BYTE>(nAddress & 0xFF), 0x20, 0xA2, byValue, 0x00, 0x20};
    DWORD dwWritten{0};
    bool bSuccess{false};
    if (dwTimeout)
#pragma warning(suppress: 26472)
      bSuccess = WriteFileTimeout(writeRequest.data(), static_cast<DWORD>(writeRequest.size()), &dwWritten, dwTimeout);
    else
#pragma warning(suppress: 26472)
      bSuccess = WriteFile(*this, writeRequest.data(), static_cast<DWORD>(writeRequest.size()), &dwWritten, nullptr);
    if (!bSuccess)
      return false;

    //Retrieve and verify the response from the HID input reports
    std::vector<BYTE> response{9, std::allocator<BYTE>{}};
    if (dwTimeout)
      bSuccess = ReadInputReportTimeout(response, dwTimeout);
    else
      bSuccess = ReadInputReport(response);
    if (!bSuccess)
      return false;
    if (response.size() != 9)
    {
      SetLastError(ERROR_INVALID_DATA);
      return false;
    }
    for (int i{1}; i<9; i++)
    {
#pragma warning(suppress: 26446)
      if (response[i] != 0xA5)
      {
        SetLastError(ERROR_INVALID_DATA);
        return false;
      }
    }
    return true;
  }

//Member variables
  OVERLAPPED m_Overlapped;
  ATL::CEvent m_ReadEvent;
};

#endif //#ifndef __FOSTATIONWRAPPERS_H__
